// Copyright © 2019-2020 Nicolas Mailhot <nim@fedoraproject.org>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

package main

import (
	"errors"
	"fmt"
	"os"

	"github.com/spf13/cobra"
	"pagure.io/modist/filter"
	"pagure.io/modist/index"
	"pagure.io/modist/install"
	"pagure.io/modist/list"
	"pagure.io/modist/logutil"
	"pagure.io/modist/pack"
	"pagure.io/modist/version"
)

func check(err error) {
	if err != nil {
		fmt.Fprintf(os.Stderr, "Error: %v\n", err)
		os.Exit(1)
	}
}

// nilIfEmpty returns nil if a string slice is empty
func nilIfEmpty(slice []string, err error) ([]string, error) {
       if len(slice) == 0 {
               return nil, err
       }
       return slice, err
}

var ErrUnknownOutputFormat = errors.New("unknown output format")

const filterHelp = `
⚠ Filtering module content

The command will try to select sensible module content by default.

--include… and --ignore… filtering options are provided to fine-tune the default
selections. --include… options are processed first, followed by --ignore…
options (this may change in the future).

Filtering options apply to relative file paths within a module tree.`

const versionHelp = `
⚠ Go module versioning

Go module versioning has no concept of post-version identifiers:
 – PRE will identify a state anterior to VERSION,
 — COMMIT alone will identify a state anterior to any pre-version of VERSION,
 — COMMIT and PRE will identify a state, anterior to VERSION, preceded by a
   series of pre-versions, ending in PRE.

When processing COMMITs and PREs, take care to increment VERSION so it
identifies the first version that will be released after those states.

⚠ RPM versioning

Since RPM versions can end up in filenames on case-insensitive filesystems or
file servers, modist will refuse to honor PRE ordering that relies on
casing.`

const brHelp = `
⚠ Go module requires and build requires differences

A Go module can declare replacements and exclusions that only apply to the
module own build. Therefore a module build requirements, may differ from the
requirements it exports to other modules.

This difference may cause problems, if other modules attempt to build the code
of the original module, with incompatible requirements, that the module masked
in its own build.

modist computes the build requirements of a synchronized set of modules with
the same versionning. Attempting to build non-synchronized sets of
modules in a single pass is not supported.`

var flagParams = map[string]func(cmd *cobra.Command){
	"version": func(cmd *cobra.Command) {
		cmd.Flags().
			StringP("version", "v", "", "use VERSION as semantic version core string")
	},
	"pre-release": func(cmd *cobra.Command) {
		cmd.Flags().
			StringP("pre-release", "P", "", "use PRE as semantic version pre-release string (PRE should not include VERSION in any form)")
	},
	"commit": func(cmd *cobra.Command) {
		cmd.Flags().StringP("commit", "C", "", "use COMMIT identifier (usually, a hash)")
	},
	"time": func(cmd *cobra.Command) {
		cmd.Flags().
			StringP("time", "H", "", `use source TIMESTAMP, in RFC 3339 format (as produced by “date -u --iso-8601=seconds”)`)
	},
	"build": func(cmd *cobra.Command) {
		cmd.Flags().
			StringP("build", "B", "", "use BUILD as semantic version build string (release in RPM conventions)")
	},
	"epoch": func(cmd *cobra.Command) {
		cmd.Flags().IntP("epoch", "Z", 0, "use versioning EPOCH")
	},
	"include-module": func(cmd *cobra.Command) {
		cmd.Flags().
			StringSliceP("include-module", "M", nil, "process this module if not set, will process all found modules")
	},
	"ignore-module": func(cmd *cobra.Command) {
		cmd.Flags().StringSliceP("ignore-module", "m", nil, "ignore this module")
	},
	"only-name": func(cmd *cobra.Command) {
		cmd.Flags().BoolP("only-name", "n", false, "limit report to the module name")
	},
	"only-version": func(cmd *cobra.Command) {
		cmd.Flags().Bool("only-version", false, "limit report to the module version")
	},
	"only-root": func(cmd *cobra.Command) {
		cmd.Flags().BoolP("only-root", "r", false, "limit report to the module file path")
	},
	"output": func(cmd *cobra.Command) {
		cmd.Flags().
			StringP("output", "o", "", "report the result of the command to output file the report is not affected by DESTDIR or PREFIX")
	},
	"format-rpm": func(cmd *cobra.Command) {
		cmd.Flags().StringP("format", "O", "rpm", "use the selected output format")
	},
	"format-txt": func(cmd *cobra.Command) {
		cmd.Flags().StringP("format", "O", "txt", "use the selected output format")
	},
	"include-file": func(cmd *cobra.Command) {
		cmd.Flags().
			StringSliceP("include-file", "F", nil, "include files with an exact path match")
	},
	"include-extension": func(cmd *cobra.Command) {
		cmd.Flags().
			StringSliceP("include-extension", "E", nil, "include files ending in this extension")
	},
	"include-directory": func(cmd *cobra.Command) {
		cmd.Flags().
			StringSliceP("include-directory", "D", nil, "include files contained in the directory itself do not include files contained in its subdirectories")
	},
	"include-tree": func(cmd *cobra.Command) {
		cmd.Flags().
			StringSliceP("include-tree", "T", nil, "include files contained in the tree root and all its subdirectories")
	},
	"include-regexp": func(cmd *cobra.Command) {
		cmd.Flags().
			StringSliceP("include-regexp", "R", nil, "include files matching this regexp")
	},
	"ignore-file": func(cmd *cobra.Command) {
		cmd.Flags().
			StringSliceP("ignore-file", "f", nil, "ignore files with an exact path match")
	},
	"ignore-extension": func(cmd *cobra.Command) {
		cmd.Flags().
			StringSliceP("ignore-extension", "e", nil, "ignore files ending in this extension")
	},
	"ignore-directory": func(cmd *cobra.Command) {
		cmd.Flags().
			StringSliceP("ignore-directory", "d", nil, "ignore files contained in the directory itself do not ignore files contained in its subdirectories")
	},
	"ignore-tree": func(cmd *cobra.Command) {
		cmd.Flags().
			StringSliceP("ignore-tree", "t", nil, "ignore files contained in the tree root and all its subdirectories")
	},
	"ignore-regexp": func(cmd *cobra.Command) {
		cmd.Flags().
			StringSliceP("ignore-regexp", "r", nil, "ignore files matching this regexp")
	},
	"prefix": func(cmd *cobra.Command) {
		cmd.Flags().
			StringP("prefix", "p", "", `pretend writing into DESTDIR actually write into PREFIX/DESTDIR`)
	},
	"qualify-root": func(cmd *cobra.Command) {
		cmd.Flags().
			BoolP("qualify-root", "Q", false, "qualify the module root with the module version (~/go/pkg/mod behaviour)")
	},
}

func main() {
	rootCmd := &cobra.Command{
		Use:     "modist",
		Short:   "A small utility to create and query Go (Golang) modules files.",
		Version: "0.1.1",
		PersistentPreRun: func(cmd *cobra.Command, args []string) {
			logLevel, _ := cmd.Flags().GetString("log-level")
			logutil.SetLogLevel(logLevel)
		},
	}

	rootCmd.PersistentFlags().
		StringP("log-level", "L", "info", "Set the log level (debug, info, warn, error, fatal)")

	addProvidesCmd(rootCmd)
	addRequiresCmd(rootCmd)
	addBuildRequiresCmd(rootCmd)
	addPathProvidesCmd(rootCmd)
	addPathRequiresCmd(rootCmd)
	addVersionCmd(rootCmd)
	addDiscoverCmd(rootCmd)
	addListCmd(rootCmd)
	addInstallCmd(rootCmd)
	addPackCmd(rootCmd)
	addIndexCmd(rootCmd)

	err := rootCmd.Execute()
	check(err)
}

func addProvidesCmd(rootCmd *cobra.Command) {
	providesCmd := &cobra.Command{
		Use:   "provides",
		Short: "Print Go module provides",
		Long:  `Print what the Go modules located in SOURCE… provide.` + versionHelp,
		Run: func(cmd *cobra.Command, args []string) {
			var report string
			var err error
			var ver *version.Version
			modVer, _ := cmd.Flags().GetString("version")
			modPre, _ := cmd.Flags().GetString("pre-release")
			modCommit, _ := cmd.Flags().GetString("commit")
			modTime, _ := cmd.Flags().GetString("time")
			modBuild, _ := cmd.Flags().GetString("build")
			modEpoch, _ := cmd.Flags().GetInt("epoch")
			includeMods, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-module"))
			ignoreMods, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-module"))
			onlyName, _ := cmd.Flags().GetBool("only-name")
			onlyVersion, _ := cmd.Flags().GetBool("only-version")
			output, _ := cmd.Flags().GetString("output")
			outputFormat, _ := cmd.Flags().GetString("format")
			sources := args

			if cmd.Flags().Changed("version") || cmd.Flags().Changed("pre-release") ||
				cmd.Flags().Changed("commit") {
				ver, err = version.New(
					modVer,
					modPre,
					modCommit,
					modTime,
					modBuild,
					modEpoch,
				)
				check(err)
			}
			_, _, report, err = list.Provides(
				sources,
				ver,
				includeMods,
				ignoreMods,
				onlyName,
				onlyVersion,
				output,
				outputFormat,
			)
			check(err)
			if report != "" {
				logutil.Info("Results:")
				fmt.Printf("%s", report)
			}
		},
	}

	flagParams["version"](providesCmd)
	flagParams["pre-release"](providesCmd)
	flagParams["commit"](providesCmd)
	flagParams["time"](providesCmd)
	flagParams["build"](providesCmd)
	flagParams["epoch"](providesCmd)
	flagParams["include-module"](providesCmd)
	flagParams["ignore-module"](providesCmd)
	flagParams["only-name"](providesCmd)
	flagParams["only-version"](providesCmd)
	flagParams["output"](providesCmd)
	flagParams["format-rpm"](providesCmd)

	rootCmd.AddCommand(providesCmd)
}

func addRequiresCmd(rootCmd *cobra.Command) {
	requiresCmd := &cobra.Command{
		Use:   "requires",
		Short: "Print Go module requires",
		Long:  "Print what the Go modules located in SOURCE… require.",
		Run: func(cmd *cobra.Command, args []string) {
			var report string
			includeMods, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-module"))
			ignoreMods, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-module"))
			output, _ := cmd.Flags().GetString("output")
			outputFormat, _ := cmd.Flags().GetString("format")
			sources := args

			_, _, report, err := list.Requires(
				sources,
				includeMods,
				ignoreMods,
				output,
				outputFormat,
			)
			check(err)
			if report != "" {
				logutil.Info("Results:")
				fmt.Printf("%s", report)
			}
		},
	}

	flagParams["include-module"](requiresCmd)
	flagParams["ignore-module"](requiresCmd)
	flagParams["output"](requiresCmd)
	flagParams["format-rpm"](requiresCmd)

	rootCmd.AddCommand(requiresCmd)
}

func addBuildRequiresCmd(rootCmd *cobra.Command) {
	buildRequiresCmd := &cobra.Command{
		Use:   "buildrequires",
		Short: "Print Go module build requires",
		Long:  `Print what building the Go modules located in SOURCE… requires.` + brHelp + versionHelp,
		Run: func(cmd *cobra.Command, args []string) {
			var report string
			var err error
			var ver *version.Version
			modVer, _ := cmd.Flags().GetString("version")
			modPre, _ := cmd.Flags().GetString("pre-release")
			modCommit, _ := cmd.Flags().GetString("commit")
			modTime, _ := cmd.Flags().GetString("time")
			modBuild, _ := cmd.Flags().GetString("build")
			modEpoch, _ := cmd.Flags().GetInt("epoch")
			includeMods, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-module"))
			ignoreMods, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-module"))
			output, _ := cmd.Flags().GetString("output")
			outputFormat, _ := cmd.Flags().GetString("format")
			sources := args

			if cmd.Flags().Changed("version") || cmd.Flags().Changed("pre-release") ||
				cmd.Flags().Changed("commit") {
				ver, err = version.New(
					modVer,
					modPre,
					modCommit,
					modTime,
					modBuild,
					modEpoch,
				)
				check(err)
			}
			_, _, report, err = list.BuildRequires(
				sources,
				ver,
				includeMods,
				ignoreMods,
				output,
				outputFormat,
			)
			check(err)
			if report != "" {
				logutil.Info("Results:")
				fmt.Printf("%s", report)
			}
		},
	}

	flagParams["version"](buildRequiresCmd)
	flagParams["pre-release"](buildRequiresCmd)
	flagParams["commit"](buildRequiresCmd)
	flagParams["time"](buildRequiresCmd)
	flagParams["build"](buildRequiresCmd)
	flagParams["epoch"](buildRequiresCmd)
	flagParams["include-module"](buildRequiresCmd)
	flagParams["ignore-module"](buildRequiresCmd)
	flagParams["output"](buildRequiresCmd)
	flagParams["format-rpm"](buildRequiresCmd)

	rootCmd.AddCommand(buildRequiresCmd)
}

func addPathProvidesCmd(rootCmd *cobra.Command) {
	pathProvidesCmd := &cobra.Command{
		Use:   "pathprovides",
		Short: "Print Go module GOPATH provides",
		Long:  `Print what the Go modules located in SOURCE… provide in GOPATH mode.` + versionHelp,
		Run: func(cmd *cobra.Command, args []string) {
			var report string
			var err error
			modVer, _ := cmd.Flags().GetString("version")
			modPre, _ := cmd.Flags().GetString("pre-release")
			modCommit, _ := cmd.Flags().GetString("commit")
			modTime, _ := cmd.Flags().GetString("time")
			modBuild, _ := cmd.Flags().GetString("build")
			modEpoch, _ := cmd.Flags().GetInt("epoch")
			includeMods, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-module"))
			ignoreMods, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-module"))
			includeFiles, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-file"))
			includeExt, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-extension"))
			includeDirs, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-directory"))
			includeTrees, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-tree"))
			includeRegexps, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-regexp"))
			ignoreFiles, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-file"))
			ignoreExt, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-extension"))
			ignoreDirs, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-directory"))
			ignoreTrees, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-tree"))
			ignoreRegexps, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-regexp"))
			onlyName, _ := cmd.Flags().GetBool("only-name")
			onlyVersion, _ := cmd.Flags().GetBool("only-version")
			output, _ := cmd.Flags().GetString("output")
			outputFormat, _ := cmd.Flags().GetString("format")
			sources := args

			ver, err := version.New(modVer, modPre, modCommit, modTime, modBuild, modEpoch)
			check(err)
			filters := filter.Default
			filters, err = filters.AddByType(
				includeFiles,
				includeExt,
				includeDirs,
				includeTrees,
				includeRegexps,
				ignoreFiles,
				ignoreExt,
				ignoreDirs,
				ignoreTrees,
				ignoreRegexps,
			)
			check(err)
			_, _, report, err = filter.PathProvides(
				sources,
				ver,
				includeMods,
				ignoreMods,
				filters,
				onlyName,
				onlyVersion,
				output,
				outputFormat,
			)
			check(err)
			if report != "" {
				logutil.Info("Results:")
				fmt.Printf("%s", report)
			}
		},
	}

	flagParams["version"](pathProvidesCmd)
	flagParams["pre-release"](pathProvidesCmd)
	flagParams["commit"](pathProvidesCmd)
	flagParams["time"](pathProvidesCmd)
	flagParams["build"](pathProvidesCmd)
	flagParams["epoch"](pathProvidesCmd)
	flagParams["include-module"](pathProvidesCmd)
	flagParams["ignore-module"](pathProvidesCmd)
	flagParams["include-file"](pathProvidesCmd)
	flagParams["include-extension"](pathProvidesCmd)
	flagParams["include-directory"](pathProvidesCmd)
	flagParams["include-tree"](pathProvidesCmd)
	flagParams["include-regexp"](pathProvidesCmd)
	flagParams["ignore-file"](pathProvidesCmd)
	flagParams["ignore-extension"](pathProvidesCmd)
	flagParams["ignore-directory"](pathProvidesCmd)
	flagParams["ignore-tree"](pathProvidesCmd)
	flagParams["ignore-regexp"](pathProvidesCmd)
	flagParams["only-name"](pathProvidesCmd)
	flagParams["only-version"](pathProvidesCmd)
	flagParams["output"](pathProvidesCmd)
	flagParams["format-rpm"](pathProvidesCmd)

	rootCmd.AddCommand(pathProvidesCmd)
}

func addPathRequiresCmd(rootCmd *cobra.Command) {
	pathRequiresCmd := &cobra.Command{
		Use:   "pathrequires",
		Short: "Print Go module GOPATH requires",
		Long:  "Print what the Go modules located in SOURCE… require in GOPATH mode.",
		Run: func(cmd *cobra.Command, args []string) {
			var report string
			includeMods, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-module"))
			ignoreMods, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-module"))
			output, _ := cmd.Flags().GetString("output")
			outputFormat, _ := cmd.Flags().GetString("format")
			sources := args

			_, _, report, err := list.PathRequires(
				sources,
				includeMods,
				ignoreMods,
				output,
				outputFormat,
			)
			check(err)
			if report != "" {
				logutil.Info("Results:")
				fmt.Printf("%s", report)
			}
		},
	}

	flagParams["include-module"](pathRequiresCmd)
	flagParams["ignore-module"](pathRequiresCmd)
	flagParams["output"](pathRequiresCmd)
	flagParams["format-rpm"](pathRequiresCmd)

	rootCmd.AddCommand(pathRequiresCmd)
}

func addVersionCmd(rootCmd *cobra.Command) {
	versionCmd := &cobra.Command{
		Use:   "version",
		Short: "Compute and convert version strings",
		Long:  `Compute version strings from individual components, or convert between version formats.` + versionHelp,
		Run: func(cmd *cobra.Command, args []string) {
			modVer, _ := cmd.Flags().GetString("version")
			modPre, _ := cmd.Flags().GetString("pre-release")
			modCommit, _ := cmd.Flags().GetString("commit")
			modTime, _ := cmd.Flags().GetString("time")
			modBuild, _ := cmd.Flags().GetString("build")
			modEpoch, _ := cmd.Flags().GetInt("epoch")
			outputFormat, _ := cmd.Flags().GetString("format")

			ver, err := version.New(modVer, modPre, modCommit, modTime, modBuild, modEpoch)
			check(err)
			result, err := ver.Format(outputFormat)
			check(err)
			fmt.Printf("%s\n", result)
		},
	}

	flagParams["version"](versionCmd)
	flagParams["pre-release"](versionCmd)
	flagParams["commit"](versionCmd)
	flagParams["time"](versionCmd)
	flagParams["build"](versionCmd)
	flagParams["epoch"](versionCmd)
	flagParams["format-rpm"](versionCmd)

	rootCmd.AddCommand(versionCmd)
}

func addDiscoverCmd(rootCmd *cobra.Command) {
	discoverCmd := &cobra.Command{
		Use:   "discover",
		Short: "Discover unpacked Go modules",
		Long:  `Discover the Go modules located in SOURCE…` + filterHelp,
		Run: func(cmd *cobra.Command, args []string) {
			var report string
			includeMods, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-module"))
			ignoreMods, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-module"))
			onlyRoot, _ := cmd.Flags().GetBool("only-root")
			output, _ := cmd.Flags().GetString("output")
			outputFormat, _ := cmd.Flags().GetString("format")
			sources := args

			_, _, report, err := list.Discover(
				sources,
				includeMods,
				ignoreMods,
				onlyRoot,
				output,
				outputFormat,
			)
			check(err)
			if report != "" {
				logutil.Info("Results:")
				fmt.Printf("%s", report)
			}
		},
	}

	flagParams["include-module"](discoverCmd)
	flagParams["ignore-module"](discoverCmd)
	flagParams["only-root"](discoverCmd)
	flagParams["output"](discoverCmd)
	flagParams["format-txt"](discoverCmd)

	rootCmd.AddCommand(discoverCmd)
}

func addListCmd(rootCmd *cobra.Command) {
	listCmd := &cobra.Command{
		Use:   "list",
		Short: "List unpacked Go modules",
		Long:  `List the unpacked Go modules located in SOURCE…` + filterHelp,
		Run: func(cmd *cobra.Command, args []string) {
			var report string
			includeMods, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-module"))
			ignoreMods, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-module"))
			includeFiles, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-file"))
			includeExt, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-extension"))
			includeDirs, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-directory"))
			includeTrees, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-tree"))
			includeRegexps, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-regexp"))
			ignoreFiles, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-file"))
			ignoreExt, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-extension"))
			ignoreDirs, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-directory"))
			ignoreTrees, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-tree"))
			ignoreRegexps, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-regexp"))
			output, _ := cmd.Flags().GetString("output")
			outputFormat, _ := cmd.Flags().GetString("format")
			sources := args

			filters := filter.Default
			filters, err := filters.AddByType(
				includeFiles,
				includeExt,
				includeDirs,
				includeTrees,
				includeRegexps,
				ignoreFiles,
				ignoreExt,
				ignoreDirs,
				ignoreTrees,
				ignoreRegexps,
			)
			check(err)
			_, _, report, err = filter.List(
				sources,
				nil,
				includeMods,
				ignoreMods,
				filters,
				output,
				outputFormat,
				true,
			)
			check(err)
			if report != "" {
				logutil.Info("Results:")
				fmt.Printf("%s", report)
			}
		},
	}

	flagParams["include-module"](listCmd)
	flagParams["ignore-module"](listCmd)
	flagParams["include-file"](listCmd)
	flagParams["include-extension"](listCmd)
	flagParams["include-directory"](listCmd)
	flagParams["include-tree"](listCmd)
	flagParams["include-regexp"](listCmd)
	flagParams["ignore-file"](listCmd)
	flagParams["ignore-extension"](listCmd)
	flagParams["ignore-directory"](listCmd)
	flagParams["ignore-tree"](listCmd)
	flagParams["ignore-regexp"](listCmd)
	flagParams["output"](listCmd)
	flagParams["format-txt"](listCmd)

	rootCmd.AddCommand(listCmd)
}

func addInstallCmd(rootCmd *cobra.Command) {
	installCmd := &cobra.Command{
		Use:   "install",
		Short: "Install unpacked Go modules",
		Long:  `Install the unpacked Go modules located in SOURCE… into DESTDIR.` + filterHelp + versionHelp,
		Run: func(cmd *cobra.Command, args []string) {
			var report string
			var err error
			modVer, _ := cmd.Flags().GetString("version")
			modPre, _ := cmd.Flags().GetString("pre-release")
			modCommit, _ := cmd.Flags().GetString("commit")
			modTime, _ := cmd.Flags().GetString("time")
			modBuild, _ := cmd.Flags().GetString("build")
			modEpoch, _ := cmd.Flags().GetInt("epoch")
			includeMods, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-module"))
			ignoreMods, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-module"))
			includeFiles, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-file"))
			includeExt, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-extension"))
			includeDirs, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-directory"))
			includeTrees, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-tree"))
			includeRegexps, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-regexp"))
			ignoreFiles, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-file"))
			ignoreExt, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-extension"))
			ignoreDirs, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-directory"))
			ignoreTrees, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-tree"))
			ignoreRegexps, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-regexp"))
			prefix, _ := cmd.Flags().GetString("prefix")
			qualifyRoot, _ := cmd.Flags().GetBool("qualify-root")
			output, _ := cmd.Flags().GetString("output")
			outputFormat, _ := cmd.Flags().GetString("format")
			sources := args[:len(args)-1]
			dstDir := args[len(args)-1]

			ver, err := version.New(modVer, modPre, modCommit, modTime, modBuild, modEpoch)
			check(err)
			filters := filter.Default
			filters, err = filters.AddByType(
				includeFiles,
				includeExt,
				includeDirs,
				includeTrees,
				includeRegexps,
				ignoreFiles,
				ignoreExt,
				ignoreDirs,
				ignoreTrees,
				ignoreRegexps,
			)
			check(err)
			_, report, err = install.Scan(
				sources,
				ver,
				includeMods,
				ignoreMods,
				filters,
				prefix,
				qualifyRoot,
				dstDir,
				output,
				outputFormat,
			)
			check(err)
			if report != "" {
				logutil.Info("Results:")
				fmt.Printf("%s", report)
			}
		},
	}

	flagParams["version"](installCmd)
	flagParams["pre-release"](installCmd)
	flagParams["commit"](installCmd)
	flagParams["time"](installCmd)
	flagParams["build"](installCmd)
	flagParams["epoch"](installCmd)
	flagParams["include-module"](installCmd)
	flagParams["ignore-module"](installCmd)
	flagParams["include-file"](installCmd)
	flagParams["include-extension"](installCmd)
	flagParams["include-directory"](installCmd)
	flagParams["include-tree"](installCmd)
	flagParams["include-regexp"](installCmd)
	flagParams["ignore-file"](installCmd)
	flagParams["ignore-extension"](installCmd)
	flagParams["ignore-directory"](installCmd)
	flagParams["ignore-tree"](installCmd)
	flagParams["ignore-regexp"](installCmd)
	flagParams["prefix"](installCmd)
	flagParams["qualify-root"](installCmd)
	flagParams["output"](installCmd)
	flagParams["format-rpm"](installCmd)

	rootCmd.AddCommand(installCmd)
}

func addPackCmd(rootCmd *cobra.Command) {
	packCmd := &cobra.Command{
		Use:   "pack",
		Short: "Pack unpacked Go modules",
		Long:  `Pack the unpacked Go modules located in SOURCE… into DESTDIR.` + filterHelp + versionHelp,
		Run: func(cmd *cobra.Command, args []string) {
			var report string
			var err error
			modVer, _ := cmd.Flags().GetString("version")
			modPre, _ := cmd.Flags().GetString("pre-release")
			modCommit, _ := cmd.Flags().GetString("commit")
			modTime, _ := cmd.Flags().GetString("time")
			modBuild, _ := cmd.Flags().GetString("build")
			modEpoch, _ := cmd.Flags().GetInt("epoch")
			includeMods, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-module"))
			ignoreMods, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-module"))
			includeFiles, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-file"))
			includeExt, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-extension"))
			includeDirs, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-directory"))
			includeTrees, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-tree"))
			includeRegexps, _ := nilIfEmpty(cmd.Flags().GetStringSlice("include-regexp"))
			ignoreFiles, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-file"))
			ignoreExt, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-extension"))
			ignoreDirs, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-directory"))
			ignoreTrees, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-tree"))
			ignoreRegexps, _ := nilIfEmpty(cmd.Flags().GetStringSlice("ignore-regexp"))
			prefix, _ := cmd.Flags().GetString("prefix")
			output, _ := cmd.Flags().GetString("output")
			outputFormat, _ := cmd.Flags().GetString("format")
			sources := args[:len(args)-1]
			dstDir := args[len(args)-1]

			ver, err := version.New(modVer, modPre, modCommit, modTime, modBuild, modEpoch)
			check(err)
			filters := filter.Default
			filters, err = filters.AddByType(
				includeFiles,
				includeExt,
				includeDirs,
				includeTrees,
				includeRegexps,
				ignoreFiles,
				ignoreExt,
				ignoreDirs,
				ignoreTrees,
				ignoreRegexps,
			)
			check(err)
			_, report, err = pack.Scan(
				sources,
				ver,
				includeMods,
				ignoreMods,
				filters,
				prefix,
				dstDir,
				output,
				outputFormat,
			)
			check(err)
			if report != "" {
				logutil.Info("Results:")
				fmt.Printf("%s", report)
			}
		},
	}

	flagParams["version"](packCmd)
	flagParams["pre-release"](packCmd)
	flagParams["commit"](packCmd)
	flagParams["time"](packCmd)
	flagParams["build"](packCmd)
	flagParams["epoch"](packCmd)
	flagParams["include-module"](packCmd)
	flagParams["ignore-module"](packCmd)
	flagParams["include-file"](packCmd)
	flagParams["include-extension"](packCmd)
	flagParams["include-directory"](packCmd)
	flagParams["include-tree"](packCmd)
	flagParams["include-regexp"](packCmd)
	flagParams["ignore-file"](packCmd)
	flagParams["ignore-extension"](packCmd)
	flagParams["ignore-directory"](packCmd)
	flagParams["ignore-tree"](packCmd)
	flagParams["ignore-regexp"](packCmd)
	flagParams["prefix"](packCmd)
	flagParams["output"](packCmd)
	flagParams["format-rpm"](packCmd)

	rootCmd.AddCommand(packCmd)
}

func addIndexCmd(rootCmd *cobra.Command) {
	indexCmd := &cobra.Command{
		Use:   "index",
		Short: "Index a GOPROXY Go module repository",
		Long: `Index a repository of Go modules that conforms to
https://tip.golang.org/cmd/go/#hdr-Module_proxy_protocol`,
		Run: func(cmd *cobra.Command, args []string) {
			dstDir := args[0]
			_, err := index.Goproxy(dstDir)
			check(err)
		},
	}

	rootCmd.AddCommand(indexCmd)
}
