// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.
//
// SPDX-License-Identifier: GPL-3.0-or-later

package descriptor

import (
	"errors"
	"fmt"
	"io"
	"os"
	"path/filepath"
	"regexp"
	"sort"
	"strings"

	"golang.org/x/mod/modfile"
	"golang.org/x/mod/module"
	"golang.org/x/mod/semver"
	"pagure.io/modist/logutil"
	"pagure.io/modist/version"
)

var (
	ErrNoDescriptor      = errors.New("no readable module descriptor")
	ErrBadDescriptor     = errors.New("no parseable module descriptor")
	ErrBadExt            = errors.New("bad file extension")
	ErrDuplicateRequires = errors.New(
		"the required module path has already been declared",
	)
	ErrUnknownOutputFormat = errors.New("unknown output format")
)

type Descriptor struct {
	*modfile.File
}

// New parses a Go .mod module descriptor without using the “go mod” command
func New(path string) (*Descriptor, error) {
	path = filepath.Clean(path)
	if filepath.Ext(path) != ".mod" {
		return nil, ErrBadExt
	}
	file, err := os.Open(path)
	if err != nil {
		return nil, ErrNoDescriptor
	}
	defer file.Close()
	rawdata, err := io.ReadAll(file)
	if err != nil {
		return nil, ErrNoDescriptor
	}
	data, err := modfile.Parse(path, rawdata, nil)
	if err != nil {
		return &Descriptor{data}, ErrBadDescriptor
	}
	return &Descriptor{data}, nil
}

func mod4text(mod module.Version) string {
	result := mod.Path
	if mod.Version != "" {
		result += " " + mod.Version
	}
	return result
}

// Trim removes module directives that make no sense in a clean system component environment
func (descriptor *Descriptor) Trim() error {
	logutil.Info(" – trimming     %s\n", descriptor.Path())

	// Remove local replaces to force the use of the system component version and warn on
	// others
	for _, r := range descriptor.Replace {
		local, err := regexp.MatchString(`^(\.\.?)?/`, r.New.Path)
		if err != nil {
			return fmt.Errorf("could not check if %s is local, %v", r.New.Path, err)
		}
		if local {
			logutil.Info(
				"     removing local replacement of %s, please package it separately\n",
				mod4text(r.Old),
			)

			if err := descriptor.DropReplace(r.Old.Path, r.Old.Version); err != nil {
				return fmt.Errorf(
					"could not remove the replacement of  %s, %v",
					mod4text(r.Old),
					err,
				)
			}
		} else {
			logutil.Warn("Replacement of %s with %s\n", mod4text(r.Old), mod4text(r.New))
			logutil.Warn("Requires and BuildRequires may differ: please fix import statements\n")
		}
	}

	// Remove indirect requires, the system package manager will resolve them at actual
	// install time
	for _, r := range descriptor.Require {
		if r.Indirect {
			logutil.Info("     removing indirect dependency on %s\n", mod4text(r.Mod))

			if err := descriptor.DropRequire(r.Mod.Path); err != nil {
				return fmt.Errorf(
					"could not remove the indirect dependency on %s, %v",
					mod4text(r.Mod),
					err,
				)
			}
		}
	}

	descriptor.Cleanup()
	return nil
}

// Path returns the module path
func (descriptor *Descriptor) Path() string {
	return descriptor.Module.Mod.Path
}

// Provides maps module paths to versions
type Provides map[string]*version.Version

// Provides returns the provides associated with a module descriptor
func (descriptor *Descriptor) Provides(
	path string,
	ver *version.Version,
) (Provides, error) {
	results := make(Provides)
	mp := descriptor.Path()
	results[mp] = ver
	if results[mp] != nil {
		if err := module.Check(mp, results[mp].Go()); err != nil {
			return results, err
		}
	}
	return results, nil
}

// RPM formats the Provides map in RPM format
func (provides Provides) RPM(onlyName bool, onlyVersion bool, namespace string) string {
	var resultlist []string
	var provs []string
	for prov := range provides {
		provs = append(provs, prov)
	}
	sort.Strings(provs)
	for _, prov := range provs {
		line := ""
		sep := ""
		if !onlyName && provides[prov] != nil {
			line = provides[prov].RPM()
			sep = " = "
		}
		if !onlyVersion {
			line = namespace + "(" + prov + ")" + sep + line
		}
		resultlist = append(resultlist, line)
	}
	sort.Strings(resultlist)
	return joinWithNewLine(resultlist)
}

// Go formats the Provides map in Go format
func (provides Provides) Go(onlyName bool, onlyVersion bool) string {
	var resultlist []string
	var provs []string
	for prov := range provides {
		provs = append(provs, prov)
	}
	sort.Strings(provs)
	for _, prov := range provs {
		line := ""
		sep := ""
		if !onlyName && provides[prov] != nil {
			line = provides[prov].Go()
			sep = " "
		}
		if !onlyVersion {
			line = prov + sep + line
		}
		resultlist = append(resultlist, line)
	}
	sort.Strings(resultlist)
	return joinWithNewLine(resultlist)
}

// Format formats the Provides map based on the specified format
func (provides Provides) Format(
	format string,
	onlyName bool,
	onlyVersion bool,
	namespace string,
) (string, error) {
	switch format {
	case "rpm":
		return provides.RPM(onlyName, onlyVersion, namespace), nil
	case "go":
		return provides.Go(onlyName, onlyVersion), nil
	}
	return "", ErrUnknownOutputFormat
}

// VersionRange represents a version range with floor and excludes
type VersionRange struct {
	Floor    *version.Version   `json:",omitempty"`
	Excludes []*version.Version `json:",omitempty"`
}

// Clean cleans the VersionRange by sorting and deduplicating the excludes
func (vr *VersionRange) Clean() {
	exs := vr.Excludes
	sort.Slice(exs, func(i, j int) bool {
		return semver.Compare(exs[i].Go(), exs[j].Go()) == -1
	})
	vr.Excludes = uniqueVersionList(vr.Excludes, vr.Floor)
}

// Requires maps module paths to version ranges
type Requires map[string]*VersionRange

// AnyRequires returns the requires associated with a module descriptor
func (descriptor *Descriptor) AnyRequires(build bool) (Requires, error) {
	results := make(Requires)
	if err := descriptor.Trim(); err != nil {
		return results, err
	}
	replaces := make(map[module.Version]module.Version)
	if build {
		for _, replace := range descriptor.Replace {
			replaces[replace.Old] = replace.New
		}
	}
	excludes := make(map[string]map[string]struct{})
	if build {
		for _, exclude := range descriptor.Exclude {
			if _, exists := excludes[exclude.Mod.Path]; !exists {
				excludes[exclude.Mod.Path] = make(map[string]struct{})
			}
			excludes[exclude.Mod.Path][exclude.Mod.Version] = struct{}{}
		}
	}
	for _, require := range descriptor.Require {
		m := resolveModule(require.Mod, replaces)
		if _, exists := results[m.Path]; exists {
			return results, ErrDuplicateRequires
		}
		var vr VersionRange
		results[m.Path] = &vr
		var err error
		vr.Floor, err = version.FromGo(m.Version)
		if err != nil {
			return results, err
		}
		if _, exists := excludes[m.Path]; exists {
			for ex := range excludes[m.Path] {
				exclude, err := version.FromGo(ex)
				if err != nil {
					return results, err
				}
				vr.Excludes = append(vr.Excludes, exclude)
			}
			vr.Clean()
		}
	}
	return results, nil
}

// Requires returns the runtime requirements associated with a module descriptor
func (descriptor *Descriptor) Requires() (Requires, error) {
	return descriptor.AnyRequires(false)
}

// BuildRequires returns the build requirements associated with a module descriptor
func (descriptor *Descriptor) BuildRequires() (Requires, error) {
	return descriptor.AnyRequires(true)
}

// RPM formats the Requires map in RPM format
func (requires Requires) RPM(namespace string) string {
	var resultlist []string
	var reqs []string
	for req := range requires {
		reqs = append(reqs, req)
	}
	sort.Strings(reqs)
	for _, req := range reqs {
		rp := namespace + "(" + req + ")"
		line := rp
		if requires[req].Floor != nil {
			line = line + " >= " + requires[req].Floor.RPM()
		}
		line = formatWithExcludes(line, requires[req].Excludes, rp)
		resultlist = append(resultlist, line)
	}
	sort.Strings(resultlist)
	return joinWithNewLine(resultlist)
}

// Go formats the Requires map in Go format
func (requires Requires) Go() string {
	var resultlist []string
	var reqs []string
	for req := range requires {
		reqs = append(reqs, req)
	}
	sort.Strings(reqs)
	for _, req := range reqs {
		line := req
		if requires[req].Floor != nil {
			line = line + " " + requires[req].Floor.Go()
		}
		if len(requires[req].Excludes) > 0 {
			line = line + " excluding:"
			for _, ex := range requires[req].Excludes {
				line = line + " " + ex.Go()
			}
		}
		resultlist = append(resultlist, line)
	}
	sort.Strings(resultlist)
	return joinWithNewLine(resultlist)
}

// Format formats the Requires map based on the specified format
func (requires Requires) Format(format string, namespace string) (string, error) {
	switch format {
	case "rpm":
		return requires.RPM(namespace), nil
	case "go":
		return requires.Go(), nil
	}
	return "", ErrUnknownOutputFormat
}

// Helper functions

// joinWithNewLine joins strings with a newline separator
func joinWithNewLine(items []string) string {
	return strings.Join(items, "\n")
}

// resolveModule resolves the module version using the replaces map.
func resolveModule(
	m module.Version,
	replaces map[module.Version]module.Version,
) module.Version {
	if r, exists := replaces[m]; exists {
		return r
	}
	if r, exists := replaces[module.Version{Path: m.Path}]; exists {
		return r
	}
	return m
}

// uniqueVersionList returns a unique list of versions excluding duplicates and sorting
// them.
func uniqueVersionList(
	excludes []*version.Version,
	floor *version.Version,
) []*version.Version {
	var unique []*version.Version
	for i, ex := range excludes {
		if semver.Compare(floor.Go(), ex.Go()) == -1 {
			if i == 0 || semver.Compare(excludes[i-1].Go(), ex.Go()) == -1 {
				unique = append(unique, ex)
			}
		}
	}
	return unique
}

// formatWithExcludes formats a string with excludes in RPM format.
func formatWithExcludes(line string, excludes []*version.Version, rp string) string {
	var exs []string
	for i, ex := range excludes {
		exv := ex.RPM()
		if i > 0 {
			exs = append(exs, " or ")
		}
		exs = append(exs, rp+" = "+exv)
	}
	if len(exs) > 1 {
		line += " without (" + strings.Join(exs, "") + ")"
	} else if len(exs) == 1 {
		line += " without " + exs[0]
	}
	return line
}
