// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.
//
// SPDX-License-Identifier: GPL-3.0-or-later

package files

import (
	"os"
	"path/filepath"
	"syscall"
	"time"

	"pagure.io/modist/descriptor"
	"pagure.io/modist/logutil"
	"pagure.io/modist/modfiles"
	"pagure.io/modist/version"
)

type Files struct {
	// The module root directory fileystem path
	Root string
	// The module version
	Version *version.Version
	// The module descriptor
	Descriptor *descriptor.Descriptor `json:"-"`
	// The list of the relative module paths of the files that will be included
	Selected []string `json:",omitempty"`
	// The list of the relative module paths of the files ignored after filtering
	Ignored []string `json:",omitempty"`
}

// New converts a modfiles.ModFiles structure to a Files structure, without changing path
// state
func New(module *modfiles.ModFiles) *Files {
	converted := &Files{
		Root:       module.Root,
		Version:    module.Version,
		Descriptor: module.Descriptor,
	}
	for path, selected := range module.Files {
		if selected {
			converted.Selected = append(converted.Selected, path)
		} else {
			converted.Ignored = append(converted.Ignored, path)
		}
	}
	return converted
}

// ModFiles converts back to a modfiles.ModFiles structure when needed
func (module *Files) ModFiles() *modfiles.ModFiles {
	converted := &modfiles.ModFiles{
		Root:       module.Root,
		Version:    module.Version,
		Descriptor: module.Descriptor,
		Files:      make(map[string]bool),
	}
	for _, path := range module.Selected {
		converted.Files[path] = true
	}
	for _, path := range module.Ignored {
		converted.Files[path] = false
	}
	return converted
}

// MTime returns the most recent modification time of the selected files within a module
func (module *Files) MTime() time.Time {
	var latestModTime time.Time
	for _, file := range module.Selected {
		if !isIgnoredFile(file) {
			modTime := getFileModTime(filepath.Join(module.Root, file))
			if modTime.After(latestModTime) {
				latestModTime = modTime
			}
		}
	}
	return latestModTime
}

// getFileModTime returns the modification time of the specified file
func getFileModTime(filePath string) time.Time {
	info, err := os.Lstat(filePath)
	if err != nil {
		logutil.Warn("Error stating file %s: %v", filePath, err)
		return time.Time{}
	}
	stat, ok := info.Sys().(*syscall.Stat_t)
	if !ok {
		logutil.Warn("Error getting stat for file %s", filePath)
		return time.Time{}
	}
	sec, nsec := stat.Mtim.Unix()
	return time.Unix(sec, nsec).UTC()
}

// isIgnoredFile checks if a file should be ignored for modification time updates
func isIgnoredFile(filename string) bool {
	switch filename {
	case "go.mod", "go.info", "go.sum", "glide.yaml", "glide.lock":
		return true
	default:
		return false
	}
}

// UpdateTimestamp changes module timestamp to the module mtime if necessary
func (module *Files) UpdateTimestamp() {
	if module.Version != nil && module.Version.Time.IsZero() {
		module.Version.Time = module.MTime()
	}
}

func (module *Files) Path() string {
	return module.Descriptor.Path()
}

// PathProvides returns the provides associated with the module paths
func (module *Files) PathProvides() descriptor.Provides {
	pathProvides := make(descriptor.Provides)
	pathProvides[module.Path()] = module.Version
	for _, file := range module.Selected {
		dir := filepath.Join(module.Path(), filepath.Dir(file))
		pathProvides[dir] = module.Version
	}
	return pathProvides
}
