// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.
//
// SPDX-License-Identifier: GPL-3.0-or-later

package filter

import (
	"encoding/json"
	"errors"
	"fmt"
	"path/filepath"
	"regexp"
	"sort"
	"strings"
	"time"

	"pagure.io/modist/descriptor"
	"pagure.io/modist/files"
	"pagure.io/modist/list"
	"pagure.io/modist/logutil"
	"pagure.io/modist/modfiles"
	"pagure.io/modist/version"
	"pagure.io/modist/write"
)

// A module content filter
type Filter struct {
	Match  func(path string) bool
	Select bool // Select or ignore paths that match
}

type Filters []Filter

var (
	ErrUnknownFilter       = errors.New("unknown filter kind")
	ErrUnknownOutputFormat = errors.New("unknown output format")
)

// New creates a new filter based on the specified kind and expression
func New(include bool, kind string, expr string) (Filter, error) {
	filter := Filter{Select: include}
	var err error
	switch kind {
	case "file":
		filter.Match = func(path string) bool { return path == filepath.Clean(expr) }
	case "extension":
		filter.Match = func(path string) bool { return filepath.Ext(path) == "."+strings.TrimPrefix(expr, ".") }
	case "directory":
		filter.Match = func(path string) bool { return filepath.Dir(path) == filepath.Clean(expr) }
	case "tree":
		filter.Match = func(path string) bool {
			rel, err := filepath.Rel(expr, path)
			return err == nil && !strings.HasPrefix(rel, "../")
		}
	case "regexp":
		var re *regexp.Regexp
		re, err = regexp.Compile(expr)
		if err == nil {
			filter.Match = func(path string) bool { return re.MatchString(path) }
		}
	default:
		err = ErrUnknownFilter
	}
	return filter, err
}

// Add adds multiple filters to the existing Filters
func (filters Filters) Add(include bool, kind string, exprs []string) (Filters, error) {
	results := filters
	for _, expr := range exprs {
		filter, err := New(include, kind, expr)
		if err != nil {
			return results, err
		}
		results = append(results, filter)
	}
	return results, nil
}

// AddByType adds filters categorized by type to the existing Filters
func (filters Filters) AddByType(
	includeFiles, includeExtensions, includeDirectories, includeTrees, includeRegexps,
	ignoreFiles, ignoreExtensions, ignoreDirectories, ignoreTrees, ignoreRegexps []string,
) (Filters, error) {
	var err error
	results := filters

	addFilters := func(include bool, kind string, exprs []string) {
		if results, err = results.Add(include, kind, exprs); err != nil {
			logutil.Warn("Error adding filters: %v", err)
		}
	}

	addFilters(true, "file", includeFiles)
	addFilters(true, "extension", includeExtensions)
	addFilters(true, "directory", includeDirectories)
	addFilters(true, "tree", includeTrees)
	addFilters(true, "regexp", includeRegexps)
	addFilters(false, "file", ignoreFiles)
	addFilters(false, "extension", ignoreExtensions)
	addFilters(false, "directory", ignoreDirectories)
	addFilters(false, "tree", ignoreTrees)
	addFilters(false, "regexp", ignoreRegexps)

	return results, err
}

var Default Filters

// init initializes the default filters
func init() {
	defaultFilters := []struct {
		include bool
		kind    string
		expr    string
	}{
		{true, "extension", "go"},
		{true, "extension", "proto"},
		{true, "extension", "json"},
		{true, "extension", "yaml"},
		{true, "extension", "md"},
		{true, "regexp", "^(.*/)?testdata/"},
		{true, "regexp", "example"},
		{false, "file", "glide.yaml"},
		{false, "file", "glide.lock"},
		// {false, "regexp", "^(.*/)?.*_test\\.go$"},
		{false, "regexp", "^(.*/)?\\..*/$"},
		{false, "regexp", "^(.*/)?\\..*"},
	}
	for _, df := range defaultFilters {
		filter, err := New(df.include, df.kind, df.expr)
		if err != nil {
			logutil.Warn("Error initializing default filter: %v", err)
		}
		Default = append(Default, filter)
	}
}

// applyFilters applies the specified filters to the given module files
func applyFilters(module *modfiles.ModFiles, filters []Filter) {
	for _, filter := range filters {
		if filter.Select {
			for path := range module.Files {
				if filter.Match(path) {
					module.Files[path] = true
				}
			}
		} else {
			for path := range module.Files {
				if filter.Match(path) {
					module.Files[path] = false
				}
			}
		}
	}
}

// Module applies filters to a modfiles.ModFiles structure, changing the input structure
func Module(module *modfiles.ModFiles, filters []Filter) *files.Files {
	applyFilters(module, filters)
	result := files.New(module)
	sort.Strings(result.Selected)
	sort.Strings(result.Ignored)
	return result
}

type Files map[string]*files.Files

// Text returns a textual representation of the filtered files
func (modules *Files) Text(selected bool) string {
	var modPaths []string
	for modPath := range *modules {
		modPaths = append(modPaths, modPath)
	}
	sort.Strings(modPaths)
	var data strings.Builder
	for _, modPath := range modPaths {
		modFiles := (*modules)[modPath]
		if selected && modFiles.Selected != nil {
			data.WriteString(fmt.Sprintf(" – retained in  %s\n", modPath))
			for _, file := range modFiles.Selected {
				data.WriteString("     " + file + "\n")
			}
		}
		if modFiles.Ignored != nil {
			data.WriteString(fmt.Sprintf(" – ignored in  %s\n", modPath))
			for _, file := range modFiles.Ignored {
				data.WriteString("     " + file + "\n")
			}
		}
	}
	return data.String()
}

// JSON returns a JSON representation of the filtered files
func (modules *Files) JSON() (string, error) {
	data, err := json.MarshalIndent(modules, "", "  ")
	if err != nil {
		return "", fmt.Errorf("failed to marshal results as JSON data, %v", err)
	}
	return string(data), nil
}

// Format returns the filtered files in the specified format
func (modules *Files) Format(format string, selected bool) (string, error) {
	switch format {
	case "txt":
		return modules.Text(selected), nil
	case "json":
		return modules.JSON()
	}
	return "", ErrUnknownOutputFormat
}

// collectModules collects and returns the list of modules from the rawModules
func collectModules(rawModules *list.Results) []string {
	var modules []string
	rawModules.RLock()
	for module := range rawModules.Modules {
		modules = append(modules, module)
	}
	sort.Strings(modules)
	rawModules.RUnlock()
	return modules
}

// Scan lists the modules found in the tree located at root, applying filters to their file
// lists, reporting the result in output with outputFormat
func Scan(
	sources []string,
	ver *version.Version,
	includeMods, ignoreMods []string,
	filters []Filter,
) (Files, error) {
	rawModules, err := list.Scan(sources, ver, includeMods, ignoreMods, false)
	if err != nil {
		return nil, err
	}
	modules := collectModules(rawModules)
	files := make(Files)
	for _, module := range modules {
		logutil.Info(" – filtering    %s", module)

		files[module] = Module(rawModules.Modules[module], filters)
		logutil.Info(" – timestamping %s", module)

		files[module].UpdateTimestamp()

		if files[module].Version != nil {
			logutil.Info("     %s", files[module].Version.Time.UTC().Format(time.RFC3339))
		}
	}
	return files, nil
}

// List scans and lists the filtered modules in the specified output format
func List(
	sources []string,
	ver *version.Version,
	includeMods []string,
	ignoreMods []string,
	filters []Filter,
	output, outputFormat string,
	selected bool,
) (Files, write.Files, string, error) {
	written := make(write.Files)
	files, err := Scan(sources, ver, includeMods, ignoreMods, filters)
	if err != nil {
		return files, written, "", err
	}
	report, err := files.Format(outputFormat, selected)
	if err != nil {
		return files, written, "", err
	}

	if output != "" {
		err = written.File(strings.NewReader(report), "", output, "txt", "", true)
	}
	return files, written, report, err
}

// PathProvides returns the provides descriptors of the filtered modules
func (modules Files) PathProvides() descriptor.Provides {
	results := make(descriptor.Provides)
	for _, mod := range modules {
		for p, v := range mod.PathProvides() {
			results[p] = v
		}
	}
	return results
}

// IPathProvides returns the provides descriptors of the filtered modules with potential
// errors
func (modules Files) IPathProvides() (descriptor.Provides, error) {
	results := make(descriptor.Provides)
	for _, mod := range modules {
		r, err := mod.ModFiles().Provides()
		if err != nil {
			return results, err
		}
		for p, v := range r {
			results[p] = v
		}
	}
	return results, nil
}

// PathProvides returns the provides descriptors of the modules in the specified output
// format
func PathProvides(
	sources []string,
	ver *version.Version,
	includeMods, ignoreMods []string,
	filters []Filter,
	onlyName, onlyVersion bool,
	output, outputFormat string,
) (Files, write.Files, string, error) {
	written := make(write.Files)
	files, err := Scan(sources, ver, includeMods, ignoreMods, filters)
	if err != nil {
		return files, written, "", err
	}
	ipathprovides, err := files.IPathProvides()
	if err != nil {
		return files, written, "", err
	}
	report1, err := ipathprovides.Format(
		outputFormat,
		onlyName,
		onlyVersion,
		"golang-ipath",
	)
	if err != nil {
		return files, written, "", err
	}

	pathprovides := files.PathProvides()
	report2, err := pathprovides.Format(outputFormat, onlyName, onlyVersion, "golang")
	if err != nil {
		return files, written, "", err
	}

	report := report1 + "\n" + report2

	if output != "" {
		err = written.File(strings.NewReader(report), "", output, "txt", "", true)
	}
	return files, written, report, err
}
