// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.
//
// SPDX-License-Identifier: GPL-3.0-or-later

package index

import (
	"errors"
	"io/fs"
	"os"
	"path/filepath"
	"sort"
	"strings"

	"golang.org/x/mod/module"
	"golang.org/x/mod/semver"
	"pagure.io/modist/descriptor"
	"pagure.io/modist/read"
	"pagure.io/modist/write"
)

var (
	ErrNotAModuleDirectory  = errors.New("the target is not a valid module directory")
	ErrInvalidModuleVersion = errors.New(
		"mismatch between the version used in the module filename, the version written in its info file, and GOPROXY encoding rules",
	)
	ErrNothingToIndex = errors.New(
		"no module to index here, the directory was removed",
	)
)

// Directory indexes the given directory and writes the result into `written`.
// If the directory is not a valid module directory, it returns an error.
func Directory(dir string, written write.Files) error {
	if filepath.Base(dir) != "@v" {
		return ErrNotAModuleDirectory
	}

	listFile := filepath.Join(dir, "list")
	infos, err := os.ReadDir(dir)
	if err != nil {
		return err
	}

	var list []string
	for _, info := range infos {
		name := info.Name()
		radical := strings.TrimSuffix(name, ".mod")
		if radical != name {
			if err := processModule(dir, radical, &list); err != nil {
				return err
			}
		}
	}

	if len(list) == 0 {
		return removeEmptyDirectory(dir, listFile)
	}

	sortVersions(list)
	return writeListFile(list, listFile, written)
}

// processModule processes a module file in the given directory, updating the list of
// versions.
// It returns an error if the module descriptor or version information is invalid.
func processModule(dir, radical string, list *[]string) error {
	_, err := descriptor.New(filepath.Join(dir, radical+".mod"))
	if err != nil {
		return err
	}
	rev, err := read.Info(filepath.Join(dir, radical+".info"))
	if err != nil {
		return err
	}
	if module.CanonicalVersion(rev.Version) != rev.Version {
		return ErrInvalidModuleVersion
	}
	encVer, err := module.EscapeVersion(rev.Version)
	if err != nil || encVer != radical {
		return ErrInvalidModuleVersion
	}
	*list = append(*list, rev.Version)

	return nil
}

// removeEmptyDirectory removes the given directory and its list file if they are empty.
// It returns an error indicating that there was nothing to index.
func removeEmptyDirectory(dir, listFile string) error {
	if err := os.Remove(listFile); err != nil {
		return err
	}
	if err := os.Remove(dir); err != nil {
		return err
	}
	return ErrNothingToIndex
}

// sortVersions sorts the list of versions in ascending order according to semantic
// versioning.
func sortVersions(list []string) {
	sort.Slice(list, func(i, j int) bool {
		return semver.Compare(list[i], list[j]) < 0
	})
}

// writeListFile writes the sorted list of versions to the list file.
// It uses the provided `written` interface to manage the file writing.
func writeListFile(list []string, listFile string, written write.Files) error {
	var data strings.Builder
	for _, ver := range list {
		encVer, _ := module.EscapeVersion(ver)
		data.WriteString(encVer + "\n")
	}
	return written.File(strings.NewReader(data.String()), "", listFile, "", "", true)
}

// goproxyWalker is a helper function for walking the directory tree.
// It calls the Directory function on directories and handles any errors appropriately.
func goproxyWalker(results write.Files, path string, info fs.DirEntry, err error) error {
	if err != nil {
		return err
	}
	if info.IsDir() {
		if err := Directory(path, results); err != nil {
			if err == ErrNothingToIndex {
				return fs.SkipDir
			}
			if err != ErrNotAModuleDirectory {
				return err
			}
		}
	}
	return nil
}

// Goproxy walks the directory tree at the given path and indexes all modules found.
// It returns the results of the indexing process.
func Goproxy(path string) (write.Files, error) {
	results := make(write.Files)
	err := filepath.WalkDir(path, func(p string, d fs.DirEntry, err error) error {
		return goproxyWalker(results, p, d, err)
	})
	return results, err
}
