// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.
//
// SPDX-License-Identifier: GPL-3.0-or-later

package install

import (
	"os"
	"path/filepath"
	"sort"

	"pagure.io/modist/files"
	"pagure.io/modist/filter"
	"pagure.io/modist/logutil"
	"pagure.io/modist/read"
	"pagure.io/modist/version"
	"pagure.io/modist/write"
)

// Module installs a module to the destination
func Module(
	module *files.Files,
	prefix string,
	qualifyRoot bool,
	destination string,
	written write.Files,
) error {
	// Create target directory
	if err := createTargetDir(prefix, destination); err != nil {
		return err
	}
	// Get module path with or without version qualification
	modPath := getModulePath(module, qualifyRoot)
	// Get all directories for the module
	mdirs := getModuleDirectories(destination, modPath)

	// Create necessary directories
	if err := createDirectories(prefix, mdirs, written); err != nil {
		return err
	}
	// Create descriptor files for the module
	if err := createDescriptorFiles(module, prefix, destination, modPath, written); err != nil {
		return err
	}

	// Get directories and directory map for selected files
	dirs, dirmap := getDirectories(destination, modPath, module)
	if err := createDirectories(prefix, dirs, written); err != nil {
		return err
	}

	// Process selected files for the module
	if err := processSelectedFiles(module, prefix, destination, modPath, written, dirs, dirmap); err != nil {
		return err
	}

	// Update timestamps for directories
	if err := updateDirectoryTimestamps(prefix, destination, modPath, dirs, module.Root); err != nil {
		return err
	}

	return nil
}

// createTargetDir creates the target directory with all necessary parent directories
func createTargetDir(prefix, destination string) error {
	return os.MkdirAll(filepath.Join(prefix, destination), 0o777)
}

// getModulePath returns the module path, optionally qualified with the version
func getModulePath(module *files.Files, qualifyRoot bool) string {
	modPath := module.Descriptor.Module.Mod.Path
	if qualifyRoot {
		modPath += "@" + module.Version.Go()
	}
	return modPath
}

// getModuleDirectories returns a sorted list of directories for the module
func getModuleDirectories(destination, modPath string) []string {
	var mdirs []string
	for root := modPath; root != "."; root = filepath.Dir(root) {
		mdirs = append(mdirs, filepath.Join(destination, root))
	}
	sort.Strings(mdirs)
	return mdirs
}

// createDirectories creates directories listed in dirs
func createDirectories(prefix string, dirs []string, written write.Files) error {
	for _, dir := range dirs {
		if err := written.Dir(prefix, dir, ""); err != nil {
			return err
		}
	}
	return nil
}

// createDescriptorFiles creates the necessary descriptor files for the module
func createDescriptorFiles(
	module *files.Files,
	prefix, destination, modPath string,
	written write.Files,
) error {
	if err := written.Descriptor(module.Descriptor, prefix, filepath.Join(destination, modPath, "go"), ""); err != nil {
		return err
	}
	return written.Info(
		module.Version,
		prefix,
		filepath.Join(destination, modPath, "go"),
		"",
	)
}

// getDirectories returns a list of directories and a map of directory existence
func getDirectories(
	destination string,
	modPath string,
	module *files.Files,
) ([]string, map[string]bool) {
	var dirs []string
	dirmap := make(map[string]bool)

	for _, path := range module.Selected {
		dir := filepath.Dir(path)
		if !dirmap[dir] {
			dirmap[dir] = true
			dirs = append(dirs, filepath.Join(destination, modPath, dir))
		}
	}

	delete(dirmap, ".")
	for dir := range dirmap {
		for root := filepath.Dir(dir); root != "."; root = filepath.Dir(root) {
			dirmap[root] = true
		}
	}

	sort.Strings(dirs)
	return dirs, dirmap
}

// processSelectedFiles processes each selected file in the module
func processSelectedFiles(
	module *files.Files,
	prefix, destination, modPath string,
	written write.Files,
	dirs []string,
	dirmap map[string]bool,
) error {
	paths := make(map[string]bool)
	for _, path := range module.Selected {
		paths[path] = true
	}

	for _, path := range module.Selected {
		spath := filepath.Join(module.Root, path)
		mpath := filepath.Join(destination, modPath)
		dpath := filepath.Join(mpath, path)

		if err := handleFile(module, prefix, spath, mpath, dpath, path, written, dirmap, paths); err != nil {
			return err
		}
	}
	return nil
}

// handleFile handles a single file, determining if it's a symlink or a regular file
func handleFile(
	module *files.Files,
	prefix, spath, mpath, dpath, path string,
	written write.Files,
	dirmap, paths map[string]bool,
) error {
	target, mt, st, err := read.Symlink(module.Root, path)
	switch err {
	case read.ErrNotASymLink:
		return handleRegularFile(prefix, spath, dpath, written)
	case read.ErrOutsideSymlink:
		logutil.Warn("%s target %s is outside %s", dpath, target, mpath)
		return written.Symlink(prefix, dpath, target, spath)
	case nil:
		return handleSymlink(
			prefix,
			spath,
			dpath,
			target,
			st,
			path,
			written,
			dirmap,
			paths,
			mt,
		)
	default:
		return err
	}
}

// handleRegularFile handles the copying of a regular file
func handleRegularFile(prefix, spath, dpath string, written write.Files) error {
	file, err := os.Open(spath)
	if err != nil {
		return err
	}
	return written.File(file, prefix, dpath, "", spath, false)
}

// handleSymlink handles the creation of a symlink
func handleSymlink(
	prefix, spath, dpath, target, st, path string,
	written write.Files,
	dirmap, paths map[string]bool,
	mt string,
) error {
	if target != st {
		logutil.Info("Simplifying %s target %s to %s", dpath, target, st)
	}
	_, existsP := paths[mt]
	_, existsD := dirmap[mt]
	if !existsP && !existsD {
		logutil.Warn("%s target %s is not included in the module", path, st)
	}
	return written.Symlink(prefix, dpath, st, spath)
}

// updateDirectoryTimestamps updates the modification timestamps for directories
func updateDirectoryTimestamps(
	prefix, destination, modPath string,
	dirs []string,
	moduleRoot string,
) error {
	sort.Sort(sort.Reverse(sort.StringSlice(dirs)))
	for _, dir := range dirs {
		if err := write.SetModTime(filepath.Join(prefix, destination, modPath, dir), filepath.Join(moduleRoot, dir)); err != nil {
			return err
		}
	}
	return nil
}

// Scan installs the modules found in the tree located at root, into destination,
// prefixed by prefix, applying filters to their file lists, reporting the
// result in output with outputFormat
func Scan(
	sources []string,
	ver *version.Version,
	includeMods []string,
	ignoreMods []string,
	filters []filter.Filter,
	prefix string,
	qualifyRoot bool,
	destination string,
	output string,
	outputFormat string,
) (write.Files, string, error) {
	written := make(write.Files)
	// Scan for modules matching the filters
	rawModules, err := filter.Scan(sources, ver, includeMods, ignoreMods, filters)
	if err != nil {
		return written, "", err
	}

	var modules []string
	for module := range rawModules {
		modules = append(modules, module)
	}
	sort.Strings(modules)

	// Install each module
	for _, module := range modules {
		logutil.Info("Installing %s", module)

		if err := Module(rawModules[module], prefix, qualifyRoot, destination, written); err != nil {
			return written, "", err
		}
	}

	// Format the installation report
	report, err := written.Format(outputFormat)
	if err != nil {
		return written, "", err
	}

	// Output the report if specified
	if output != "" {
		return written, "", written.Output(output, outputFormat)
	}

	return written, report, nil
}
