// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.
//
// SPDX-License-Identifier: GPL-3.0-or-later

package list

import (
	"encoding/json"
	"errors"
	"fmt"
	"os"
	"path/filepath"
	"sort"
	"strings"
	"sync"

	"golang.org/x/mod/module"
	"golang.org/x/mod/semver"

	"pagure.io/modist/descriptor"
	"pagure.io/modist/logutil"
	"pagure.io/modist/modfiles"
	"pagure.io/modist/read"
	"pagure.io/modist/version"
	"pagure.io/modist/write"
)

type Results struct {
	sync.RWMutex
	Modules map[string]*modfiles.ModFiles
}

// Taking only module into account: we do not allow several versions of the
// same module path within a single source directory, because a source
// directory is associated with a single project version
var (
	ErrSeenModule      = errors.New("this module was already scanned")
	ErrDuplicateModule = errors.New(
		"this module conflicts with another one that declares the same path",
	)
	ErrIgnoredModule       = errors.New("the user requested to ignore this module path")
	ErrUnknownOutputFormat = errors.New("unknown output format")
	ErrDuplicateProvides   = errors.New(
		"the provided module path has already been declared",
	)
)

// Must be called with results RLock or Lock held
// Checks if a module has already been seen in the provided results
func checkSeenModule(modPath string, path string, results *Results) error {
	other, exists := results.Modules[modPath]
	if exists {
		otherinfo, err := os.Stat(other.Root)
		if err != nil {
			return err
		}
		info, err := os.Stat(path)
		if err != nil {
			return err
		}
		if os.SameFile(otherinfo, info) {
			logutil.Warn("Already seen an identical %s at %s\n", modPath, path)
			return ErrSeenModule
		}
		logutil.Warn("%s is already declared in %s\n", modPath, path)
	}
	return nil
}

// Processes a file (path) within a module tree
func moduleWalker(
	modPath string,
	ver *version.Version,
	addUnknownModules bool,
	modules map[string]bool,
	results *Results,
	onlydescr bool,
	sourcefiles bool,
	detachsubs bool,
	walksubs bool,
	path string,
	info os.FileInfo,
	err error,
) error {
	if err != nil {
		return err
	}
	relpath, err := getRelativePath(results, modPath, path)
	if err != nil {
		return err
	}

	if info.IsDir() {
		return processDirectory(
			modPath,
			ver,
			addUnknownModules,
			modules,
			results,
			onlydescr,
			sourcefiles,
			detachsubs,
			walksubs,
			path,
			relpath,
		)
	}

	return processFile(modPath, results, onlydescr, relpath)
}

// Retrieves the relative path from the module root to the given path
func getRelativePath(results *Results, modPath, path string) (string, error) {
	results.RLock()
	defer results.RUnlock()
	return filepath.Rel(results.Modules[modPath].Root, path)
}

// Processes a directory within the module tree
func processDirectory(
	modPath string,
	ver *version.Version,
	addUnknownModules bool,
	modules map[string]bool,
	results *Results,
	onlydescr bool,
	sourcefiles bool,
	detachsubs bool,
	walksubs bool,
	path, relpath string,
) error {
	if relpath == "." {
		return nil
	}
	if detachsubs {
		return handleDetachedSubmodule(
			modPath,
			ver,
			addUnknownModules,
			modules,
			results,
			onlydescr,
			sourcefiles,
			detachsubs,
			walksubs,
			path,
			relpath,
		)
	}
	return handleNormalDirectory(results, modPath, path, onlydescr, sourcefiles, relpath)
}

// Handles detached submodules within the module tree
func handleDetachedSubmodule(
	modPath string,
	ver *version.Version,
	addUnknownModules bool,
	modules map[string]bool,
	results *Results,
	onlydescr bool,
	sourcefiles bool,
	detachsubs bool,
	walksubs bool,
	path, relpath string,
) error {
	info, err := TryModule(filepath.Join(path, "go.mod"))
	if err != nil {
		if err != descriptor.ErrNoDescriptor {
			return err
		}
		return nil
	}
	if !walksubs {
		return filepath.SkipDir
	}
	err = Descriptor(info, ver, addUnknownModules, modules, results, onlydescr, walksubs)
	if err == ErrSeenModule {
		return filepath.SkipDir
	}
	if err == ErrIgnoredModule {
		return filepath.Walk(path, func(p string, info os.FileInfo, err error) error {
			return treeWalker(
				path,
				ver,
				addUnknownModules,
				modules,
				results,
				onlydescr,
				p,
				info,
				err,
			)
		})
	}
	if err == nil {
		return filepath.SkipDir
	}
	return err
}

// Handles normal directories within the module tree
func handleNormalDirectory(
	results *Results,
	modPath, path string,
	onlydescr bool,
	sourcefiles bool,
	relpath string,
) error {
	dir := filepath.Base(relpath)
	if isIgnoredDirectory(dir) {
		if onlydescr {
			return filepath.SkipDir
		}
		if sourcefiles {
			return handleSourceFiles(path, modPath, results)
		}
	}
	if sourcefiles {
		results.Lock()
		defer results.Unlock()
		return results.Modules[modPath].Sources(path)
	}
	return nil
}

// Checks if a directory should be ignored based on its name
func isIgnoredDirectory(dir string) bool {
	return dir == "testdata" || dir == "vendor" || strings.HasPrefix(dir, ".") ||
		strings.HasPrefix(dir, "_")
}

// Processes source files within a directory
func handleSourceFiles(path, modPath string, results *Results) error {
	return filepath.Walk(path, func(p string, info os.FileInfo, err error) error {
		return moduleWalker(
			modPath,
			nil,
			false,
			nil,
			results,
			false,
			false,
			false,
			false,
			p,
			info,
			err,
		)
	})
}

// Processes a file within the module tree
func processFile(modPath string, results *Results, onlydescr bool, relpath string) error {
	if !onlydescr && relpath != "go.mod" {
		results.RLock()
		if _, exists := results.Modules[modPath].Files[relpath]; !exists {
			results.RUnlock()
			results.Lock()
			defer results.Unlock()
			if _, exists := results.Modules[modPath].Files[relpath]; !exists {
				results.Modules[modPath].Files[relpath] = false
			}
		} else {
			results.RUnlock()
		}
	}
	return nil
}

// Check if a file path is a valid module descriptor; return a modfiles.ModFiles struct if
// true
func TryModule(path string) (result modfiles.ModFiles, err error) {
	descr, err := descriptor.New(path)
	if err != nil {
		return result, err
	}
	result.Root = filepath.Dir(path)
	result.Descriptor = descr
	result.Files = make(map[string]bool)
	// We don’t care if reading the Version fails – Version info may be missing
	// in the first place
	result.Version, _ = read.Version(path)
	return result, err
}

// Walks a module tree to list its constituent files
func Module(
	root string,
	ver *version.Version,
	addUnknownModules bool,
	modules map[string]bool,
	results *Results,
	onlydescr bool,
) error {
	info, err := TryModule(filepath.Join(root, "go.mod"))
	if err != nil {
		return err
	}
	version.MergeVersions(info.Version, ver)
	return Descriptor(
		info,
		info.Version,
		addUnknownModules,
		modules,
		results,
		onlydescr,
		true,
	)
}

// Walks a module tree, identified by its modfiles.ModFiles info
func Descriptor(
	info modfiles.ModFiles,
	ver *version.Version,
	addUnknownModules bool,
	modules map[string]bool,
	results *Results,
	onlydescr bool,
	walksubs bool,
) error {
	modPath := info.Path()
	root := info.Root
	if shouldIgnoreModule(modPath, addUnknownModules, modules) {
		logutil.Info("Ignoring %s at %s\n", modPath, root)
		return ErrIgnoredModule
	}
	logutil.Info("Listing %s at %s\n", modPath, root)

	err := checkAndLockModule(modPath, root, results)
	if err != nil {
		return err
	}
	results.Modules[modPath] = &info
	if onlydescr && !walksubs {
		results.Unlock()
		return nil
	}
	err = info.Sources(root)
	if err != nil {
		results.Unlock()
		return err
	}
	if ver != nil && *ver != (version.Version{}) {
		err = validateVersion(modPath, ver)
		if err != nil {
			return err
		}
		results.Modules[modPath].Version = ver
	}
	results.Unlock()
	return filepath.Walk(root, func(path string, info os.FileInfo, err error) error {
		return moduleWalker(
			modPath,
			ver,
			addUnknownModules,
			modules,
			results,
			onlydescr,
			true,
			true,
			walksubs,
			path,
			info,
			err,
		)
	})
}

// Determines if a module should be ignored based on the provided parameters
func shouldIgnoreModule(
	modPath string,
	addUnknownModules bool,
	modules map[string]bool,
) bool {
	_, known := modules[modPath]
	return !((!known && addUnknownModules) || (known && modules[modPath]))
}

// Checks if a module has been seen and locks the results if not
func checkAndLockModule(modPath, root string, results *Results) error {
	results.RLock()
	err := checkSeenModule(modPath, root, results)
	results.RUnlock()
	if err != nil {
		return err
	}
	results.Lock()
	err = checkSeenModule(modPath, root, results)
	if err != nil {
		results.Unlock()
		return err
	}
	return nil
}

// Validates the version of a module
func validateVersion(modPath string, ver *version.Version) error {
	return module.Check(modPath, ver.Go())
}

// Process a file (path) within a source tree
func treeWalker(
	root string,
	ver *version.Version,
	addUnknownModules bool,
	modules map[string]bool,
	results *Results,
	onlydescr bool,
	path string,
	info os.FileInfo,
	err error,
) error {
	if err != nil {
		return err
	}
	if info.IsDir() {
		if root != path {
			err := Module(path, ver, addUnknownModules, modules, results, onlydescr)
			if err == nil || err == ErrSeenModule {
				return filepath.SkipDir
			}
			if err != descriptor.ErrNoDescriptor && err != ErrIgnoredModule {
				return err
			}
		}
		dir := filepath.Base(path)
		if isIgnoredDirectory(dir) {
			return filepath.SkipDir
		}
	}
	return nil
}

// Selects sources from the provided list and returns their roots and information
func selectSources(
	sources []string,
) (roots []string, infos []modfiles.ModFiles, err error) {
	srcInfos := make(map[os.FileInfo]string)
	sort.Strings(sources)
	for _, src := range sources {
		srcInfo, err := os.Stat(src)
		if err != nil {
			logutil.Info("Ignoring source %s (cannot stat)\n", src)
			break
		}
		if _, exists := srcInfos[srcInfo]; exists {
			logutil.Info("Ignoring source %s (already specified)\n", src)
		} else {
			srcInfos[srcInfo] = src
			if srcInfo.IsDir() {
				roots = append(roots, src)
			} else {
				info, err := TryModule(src)
				if err != nil {
					return roots, infos, err
				}
				infos = append(infos, info)
			}
		}
	}
	return roots, infos, nil
}

// Selects modules to include or ignore based on the provided lists
func selectModules(includeMods, ignoreMods []string) (bool, map[string]bool) {
	addUnknownModules := true
	modules := make(map[string]bool)
	if includeMods != nil {
		addUnknownModules = false
		for _, modPath := range includeMods {
			modules[modPath] = true
		}
	}
	for _, modPath := range ignoreMods {
		modules[modPath] = false
	}
	return addUnknownModules, modules
}

// Walks sources to find and read Go module source trees within
func Scan(
	sources []string,
	ver *version.Version,
	includeMods []string,
	ignoreMods []string,
	onlydescr bool,
) (*Results, error) {
	results := &Results{
		Modules: make(map[string]*modfiles.ModFiles),
	}
	roots, infos, err := selectSources(sources)
	if err != nil {
		return results, err
	}
	addUnknownModules, modules := selectModules(includeMods, ignoreMods)
	errors := make(chan error)
	for _, root := range roots {
		go func(root string) {
			logutil.Info("Searching for unpacked Go modules in %s…\n", root)
			err := Module(root, ver, addUnknownModules, modules, results, onlydescr)
			if err == descriptor.ErrNoDescriptor || err == ErrIgnoredModule {
				err = filepath.Walk(
					root,
					func(path string, info os.FileInfo, err error) error {
						return treeWalker(
							root,
							ver,
							addUnknownModules,
							modules,
							results,
							onlydescr,
							path,
							info,
							err,
						)
					},
				)
			}
			errors <- err
		}(root)
	}
	for _, info := range infos {
		go func(info modfiles.ModFiles) {
			logutil.Info("Looking at the %s module in %s…\n", info.Path(), info.Root)
			errors <- Descriptor(info, ver, addUnknownModules, modules, results, onlydescr, false)
		}(info)
	}
	for i := 0; i < len(roots)+len(infos); i++ {
		if err := <-errors; err != nil && err != ErrSeenModule {
			return results, err
		}
	}
	return results, nil
}

// Discovers Go modules within the provided sources and generates a report
func Discover(
	sources []string,
	includeMods []string,
	ignoreMods []string,
	onlyRoot bool,
	output string,
	outputFormat string,
) (*Results, write.Files, string, error) {
	written := make(write.Files)
	results, err := Scan(sources, nil, includeMods, ignoreMods, true)
	if err != nil {
		return results, written, "", err
	}
	report, err := results.Format(outputFormat, onlyRoot)
	if err != nil {
		return results, written, "", err
	}

	if output != "" {
		err = written.File(strings.NewReader(report), "", output, outputFormat, "", true)
	}
	return results, written, report, err
}

// Lists the modules in the provided format
func (modules *Results) List(separator string, onlyRoot bool) string {
	modules.RLock()
	defer modules.RUnlock()
	var resultlist []string
	for _, mod := range modules.Modules {
		result := mod.Root
		if !onlyRoot {
			result = mod.Path() + separator + mod.Version.Go() + separator + result
		}
		resultlist = append(resultlist, result)
	}
	sort.Strings(resultlist)
	return strings.Join(resultlist, "\n")
}

// Lists the modules in text format
func (modules *Results) Text(onlyRoot bool) string {
	return modules.List("\t", onlyRoot)
}

// Lists the modules in CSV format
func (modules *Results) CSV(onlyRoot bool) string {
	return modules.List(";", onlyRoot)
}

// Lists the modules in JSON format
func (modules *Results) JSON() (string, error) {
	type M struct {
		Module  string
		Version string
	}
	results := make(map[string]M)
	for _, mod := range modules.Modules {
		var m M
		m.Module = mod.Path()
		m.Version = mod.Version.Go()
		results[mod.Root] = m
	}
	data, err := json.MarshalIndent(results, "", "  ")
	if err != nil {
		return "", fmt.Errorf("failed to marshal results as JSON data: %v", err)
	}
	return string(data), nil
}

// Formats the module list in the specified format
func (modules *Results) Format(format string, onlyRoot bool) (string, error) {
	switch format {
	case "txt":
		return modules.Text(onlyRoot), nil
	case "csv":
		return modules.CSV(onlyRoot), nil
	case "json":
		return modules.JSON()
	}
	return "", ErrUnknownOutputFormat
}

// Retrieves the provides information for the modules
func (modules *Results) Provides() (descriptor.Provides, error) {
	results := make(descriptor.Provides)
	modules.RLock()
	defer modules.RUnlock()
	for _, mod := range modules.Modules {
		ps, err := mod.Provides()
		if err != nil {
			return results, err
		}
		for p := range ps {
			if _, exists := results[p]; exists {
				return results, ErrDuplicateProvides
			}
			results[p] = ps[p]
		}
	}
	return results, nil
}

// Retrieves the required dependencies for the modules
func (modules *Results) anyRequires(build bool) (descriptor.Requires, error) {
	results := make(descriptor.Requires)
	var reqs []descriptor.Requires
	modules.RLock()
	defer modules.RUnlock()
	for _, mod := range modules.Modules {
		rs, err := mod.AnyRequires(build)
		if err != nil {
			return results, err
		}
		reqs = append(reqs, rs)
	}
	merged := make(map[string]struct{})
	for _, rs := range reqs {
		for r := range rs {
			_, exists := results[r]
			if !exists {
				results[r] = rs[r]
			} else {
				merged[r] = struct{}{}
				if semver.Compare(results[r].Floor.Go(), rs[r].Floor.Go()) == -1 {
					results[r].Floor = rs[r].Floor
				}
				results[r].Excludes = append(results[r].Excludes, rs[r].Excludes...)
			}
		}
	}
	for r := range merged {
		results[r].Clean()
	}
	return results, nil
}

// Retrieves the runtime dependencies for the modules
func (modules *Results) Requires() (descriptor.Requires, error) {
	return modules.anyRequires(false)
}

// Retrieves the build dependencies for the modules
func (modules *Results) BuildRequires() (descriptor.Requires, error) {
	results := make(descriptor.Requires)
	rawbrs, err := modules.anyRequires(true)
	if err != nil {
		return results, err
	}
	provs, err := modules.Provides()
	if err != nil {
		return results, err
	}
	if len(provs) == 1 {
		return rawbrs, nil
	}
	rawbrl := make([]string, 0, len(rawbrs))
	for br := range rawbrs {
		rawbrl = append(rawbrl, br)
	}
	sort.Strings(rawbrl)
	logutil.Info("Consolidating BuildRequires…")

	for _, br := range rawbrl {
		retain := shouldRetainBuildRequire(br, provs, rawbrs, results)
		if retain {
			results[br] = rawbrs[br]
		}
	}
	return results, nil
}

// Determines if a build requirement should be retained based on the provided parameters
func shouldRetainBuildRequire(
	br string,
	provs descriptor.Provides,
	rawbrs, results descriptor.Requires,
) bool {
	retain := true
	_, exists := provs[br]
	if exists {
		refv := "v0.0.0"
		if provs[br] != nil {
			refv = provs[br].Go()
		}
		retain = false
		if semver.Compare(refv, rawbrs[br].Floor.Go()) == -1 {
			retain = true
		} else {
			for _, ex := range rawbrs[br].Excludes {
				if semver.Compare(refv, ex.Go()) == 0 {
					retain = true
					break
				}
			}
		}
		if retain {
			logutil.Warn(
				"The provided %s module does not satisfy the build requirements of other modules\n",
				br,
			)
		}
	}
	return retain
}

// Generates a report of the provides information for the modules
func Provides(
	sources []string,
	ver *version.Version,
	includeMods, ignoreMods []string,
	onlyName, onlyVersion bool,
	output, outputFormat string,
) (*Results, write.Files, string, error) {
	written := make(write.Files)
	scan, err := Scan(sources, ver, includeMods, ignoreMods, true)
	if err != nil {
		return nil, written, "", err
	}
	provides, err := scan.Provides()
	if err != nil {
		return nil, written, "", err
	}
	report, err := provides.Format(outputFormat, onlyName, onlyVersion, "golang-module")
	if err != nil {
		return nil, written, "", err
	}

	if output != "" {
		err = written.File(strings.NewReader(report), "", output, "txt", "", true)
	}
	return scan, written, report, err
}

// Retrieves the runtime dependencies for the modules and generates a report
func anyRequires(
	sources []string,
	includeMods []string,
	ignoreMods []string,
	namespace string,
	output string,
	outputFormat string,
) (descriptor.Requires, write.Files, string, error) {
	written := make(write.Files)
	scan, err := Scan(sources, nil, includeMods, ignoreMods, true)
	if err != nil {
		return nil, written, "", err
	}
	requires, err := scan.Requires()
	if err != nil {
		return nil, written, "", err
	}
	report, err := requires.Format(outputFormat, namespace)
	if err != nil {
		return nil, written, "", err
	}

	if output != "" {
		err = written.File(strings.NewReader(report), "", output, "txt", "", true)
	}
	return requires, written, report, err
}

// Retrieves the runtime dependencies for the modules and generates a report
func Requires(
	sources []string,
	includeMods []string,
	ignoreMods []string,
	output string,
	outputFormat string,
) (descriptor.Requires, write.Files, string, error) {
	return anyRequires(
		sources,
		includeMods,
		ignoreMods,
		"golang-module",
		output,
		outputFormat,
	)
}

// Retrieves the path dependencies for the modules and generates a report
func PathRequires(
	sources []string,
	includeMods []string,
	ignoreMods []string,
	output string,
	outputFormat string,
) (descriptor.Requires, write.Files, string, error) {
	return anyRequires(
		sources,
		includeMods,
		ignoreMods,
		"golang-ipath",
		output,
		outputFormat,
	)
}

// Retrieves the build dependencies for the modules and generates a report
func BuildRequires(
	sources []string,
	ver *version.Version,
	includeMods []string,
	ignoreMods []string,
	output string,
	outputFormat string,
) (descriptor.Requires, write.Files, string, error) {
	written := make(write.Files)
	scan, err := Scan(sources, ver, includeMods, ignoreMods, true)
	if err != nil {
		return nil, written, "", err
	}
	buildrequires, err := scan.BuildRequires()
	if err != nil {
		return nil, written, "", err
	}
	report, err := buildrequires.Format(outputFormat, "golang-module")
	if err != nil {
		return nil, written, "", err
	}

	if output != "" {
		err = written.File(strings.NewReader(report), "", output, "txt", "", true)
	}
	return buildrequires, written, report, err
}
