// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.
//
// SPDX-License-Identifier: GPL-3.0-or-later

package logutil

import (
	"log"
	"os"
	"strings"
)

const (
	DEBUG = iota
	INFO
	WARN
	ERROR
	FATAL
)

var (
	currentLogLevel = INFO
	logLevels       = map[string]int{
		"debug": DEBUG,
		"info":  INFO,
		"warn":  WARN,
		"error": ERROR,
		"fatal": FATAL,
	}
)

type customLogWriter struct{}

func (w customLogWriter) Write(bytes []byte) (int, error) {
	return os.Stderr.Write([]byte("# " + strings.TrimSpace(string(bytes)) + "\n"))
}

func SetLogLevel(level string) {
	log.SetOutput(customLogWriter{})
	if logLevel, ok := logLevels[level]; ok {
		currentLogLevel = logLevel
	} else {
		log.Printf("# Invalid log level: %s. Using default level 'info'.", level)
		currentLogLevel = INFO
	}
}

func Debug(format string, v ...interface{}) {
	if currentLogLevel <= DEBUG {
		log.Printf("DEBUG: "+format, v...)
	}
}

func Info(format string, v ...interface{}) {
	if currentLogLevel <= INFO {
		log.Printf("INFO: "+format, v...)
	}
}

func Warn(format string, v ...interface{}) {
	if currentLogLevel <= WARN {
		log.Printf("WARN: "+format, v...)
	}
}

func Error(format string, v ...interface{}) {
	if currentLogLevel <= ERROR {
		log.Printf("ERROR: "+format, v...)
	}
}

func Fatal(format string, v ...interface{}) {
	if currentLogLevel <= FATAL {
		log.Fatalf("FATAL: "+format, v...)
	}
}
