// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.
//
// SPDX-License-Identifier: GPL-3.0-or-later

package modfiles

import (
	"go/build"
	"path/filepath"

	"pagure.io/modist/descriptor"
	"pagure.io/modist/version"
)

// ModFiles holds raw unpacked module file information
type ModFiles struct {
	Descriptor *descriptor.Descriptor // The module descriptor
	Version    *version.Version       // The module version
	Root       string                 // The module root directory filesystem path
	Files      map[string]bool        // A map of the relative paths of the module files within the module, to whether they will be included in the packed module or not
}

// Path returns the module path
func (module *ModFiles) Path() string {
	return module.Descriptor.Path()
}

// Provides returns the provides associated with a module
func (module *ModFiles) Provides() (descriptor.Provides, error) {
	return module.Descriptor.Provides(module.Root, module.Version)
}

// AnyRequires returns any requires associated with a module
func (module *ModFiles) AnyRequires(build bool) (descriptor.Requires, error) {
	return module.Descriptor.AnyRequires(build)
}

// Requires returns the requires associated with a module
func (module *ModFiles) Requires() (descriptor.Requires, error) {
	return module.Descriptor.Requires()
}

// BuildRequires returns the build requires associated with a module
func (module *ModFiles) BuildRequires() (descriptor.Requires, error) {
	return module.Descriptor.BuildRequires()
}

// Sources lists the source files of a Go package, without using the “go list” command
func (module *ModFiles) Sources(path string) error {
	p, serr := build.ImportDir(path, build.IgnoreVendor)
	if serr != nil {
		// Directories without any Go source file are not a problem; other errors are
		if _, ok := serr.(*build.NoGoError); !ok {
			return serr
		}
	}
	return module.addFiles(p, path)
}

// addFiles adds files to the module from the specified build package and path
func (module *ModFiles) addFiles(p *build.Package, path string) error {
	fileGroups := [][]string{
		p.GoFiles, p.CgoFiles, p.IgnoredGoFiles,
		p.CFiles, p.CXXFiles, p.MFiles, p.HFiles, p.FFiles, p.SFiles,
		p.SwigFiles, p.SwigCXXFiles, p.TestGoFiles,
	}
	for _, files := range fileGroups {
		for _, f := range files {
			rel, err := filepath.Rel(module.Root, filepath.Join(path, f))
			if err != nil {
				return err
			}
			module.Files[rel] = true
		}
	}
	return nil
}
