// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.
//
// SPDX-License-Identifier: GPL-3.0-or-later

package pack

import (
	"sort"

	"pagure.io/modist/files"
	"pagure.io/modist/filter"
	"pagure.io/modist/logutil"
	"pagure.io/modist/version"
	"pagure.io/modist/write"
)

// Scan packs the modules found in the tree located at root into the destination,
// prefixed by prefix, attributing them version, applying filters to their file
// lists, reporting the result in output with outputFormat
func Scan(sources []string, ver *version.Version,
	includeMods []string, ignoreMods []string, filters []filter.Filter,
	prefix string, destination string,
	output string, outputFormat string,
) (written write.Files, report string, err error) {
	written = make(write.Files)

	// Scan for raw modules using specified filters
	rawModules, err := filter.Scan(sources, ver, includeMods, ignoreMods, filters)
	if err != nil {
		return written, "", err
	}

	// Sort the modules by their names
	modules := getSortedModules(rawModules)

	// Pack each module and add to written files
	for _, module := range modules {
		logutil.Info("Packing %s", module)
		err = packModule(written, rawModules[module], prefix, destination)
		if err != nil {
			return written, "", err
		}
	}

	// Generate and log the report of packed files
	report, err = logReport(written, outputFormat)
	if err != nil {
		return written, "", err
	}

	// Output the result to the specified destination
	if err := outputResult(written, output, outputFormat); err != nil {
		return written, "", err
	}

	return written, report, nil
}

// getSortedModules returns the sorted list of module names
func getSortedModules(rawModules filter.Files) []string {
	var modules []string
	for module := range rawModules {
		modules = append(modules, module)
	}
	sort.Strings(modules)
	return modules
}

// packModule packs a single module and adds it to the written files
func packModule(
	written write.Files,
	module *files.Files,
	prefix, destination string,
) error {
	return written.Pack(module, prefix, destination)
}

// logReport generates and returns the formatted report of the packed files
func logReport(written write.Files, outputFormat string) (string, error) {
	report, err := written.Format(outputFormat)
	if err != nil {
		return "", err
	}
	return report, nil
}

// outputResult writes the report to the specified output location
func outputResult(written write.Files, output, outputFormat string) error {
	if output != "" {
		return written.Output(output, outputFormat)
	}
	return nil
}
