// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.
//
// SPDX-License-Identifier: GPL-3.0-or-later

package read

import (
	"encoding/json"
	"errors"
	"io"
	"os"
	"path/filepath"
	"regexp"
	"strings"

	"pagure.io/modist/logutil"
	"pagure.io/modist/version"
)

var (
	ErrNotASymLink    = errors.New("not a symlink")
	ErrOutsideSymlink = errors.New("symlink pointing outside the module")
	ErrBadExt         = errors.New("bad file extension")
)

// Symlink tells if the file located at path under root is a symbolic link,
// and if it points outside the root.
func Symlink(root, path string) (string, string, string, error) {
	fullPath := filepath.Join(root, path)
	info, err := os.Lstat(fullPath)
	if err != nil {
		return "", "", "", err
	}
	if info.Mode()&os.ModeSymlink == 0 {
		return "", "", "", ErrNotASymLink
	}
	return processSymlink(root, fullPath)
}

// processSymlink processes the symbolic link and determines if it points outside the root.
func processSymlink(root, fullPath string) (string, string, string, error) {
	originalTarget, err := os.Readlink(fullPath)
	if err != nil {
		return "", "", "", err
	}
	simplifiedTarget := getSimplifiedTarget(originalTarget, fullPath)
	absTarget, err := filepath.Abs(simplifiedTarget)
	if err != nil {
		return "", "", "", err
	}
	absRoot, err := filepath.Abs(root)
	if err != nil {
		return "", "", "", err
	}
	modTarget, err := getRelativePath(absRoot, absTarget)
	if err != nil {
		return "", "", "", err
	}
	absPath, err := filepath.Abs(fullPath)
	if err != nil {
		return "", "", "", err
	}
	simplifiedTarget, err = filepath.Rel(filepath.Dir(absPath), absTarget)
	return originalTarget, modTarget, simplifiedTarget, err
}

// getSimplifiedTarget returns the simplified target path for a symbolic link.
func getSimplifiedTarget(originalTarget, fullPath string) string {
	if !strings.HasPrefix(originalTarget, "/") {
		return filepath.Join(filepath.Dir(fullPath), originalTarget)
	}
	return originalTarget
}

// getRelativePath returns the relative path from the root to the target.
// It ensures that the target does not point outside the root.
func getRelativePath(root, target string) (string, error) {
	modTarget, err := filepath.Rel(root, target)
	if err != nil || strings.HasPrefix(modTarget, "../") {
		return "", ErrOutsideSymlink
	}
	return modTarget, nil
}

// Version tries to infer a module version from the information present around the module
// descriptor path.
func Version(path string) (*version.Version, error) {
	path = filepath.Clean(path)
	if filepath.Ext(path) != ".mod" {
		return nil, ErrBadExt
	}

	if v, err := inferFromInfoFile(path); err == nil {
		return v, nil
	}

	if v, err := inferFromGoMod(path); err == nil {
		return v, nil
	}

	return inferFromQualifiedPath(path)
}

// inferFromInfoFile tries to infer the version from an associated .info file.
func inferFromInfoFile(path string) (*version.Version, error) {
	ip := strings.TrimSuffix(path, ".mod") + ".info"
	i, err := Info(ip)
	if err != nil {
		return nil, err
	}
	return i.Ver()
}

// inferFromGoMod tries to infer the version from the go.mod file name.
func inferFromGoMod(path string) (*version.Version, error) {
	r := strings.TrimSuffix(filepath.Base(path), ".mod")
	return version.FromGo(r)
}

// inferFromQualifiedPath tries to infer the version from the qualified path.
func inferFromQualifiedPath(path string) (*version.Version, error) {
	d, err := filepath.Abs(filepath.Dir(path))
	if err != nil {
		return nil, err
	}
	vRe := regexp.MustCompile(`.*@v(.*)$`)
	v := vRe.ReplaceAllString(filepath.Base(d), "v${1}")
	return version.FromGo(v)
}

// Info parses a Go .info file.
func Info(path string) (*version.Info, error) {
	path = filepath.Clean(path)
	if filepath.Ext(path) != ".info" {
		return nil, ErrBadExt
	}
	return readInfoFile(path)
}

// readInfoFile reads and unmarshals a .info file to extract version information.
func readInfoFile(path string) (*version.Info, error) {
	jsonFile, err := os.Open(path)
	if err != nil {
		return nil, logError("Error opening JSON file", err)
	}
	defer jsonFile.Close()

	byteValue, err := io.ReadAll(jsonFile)
	if err != nil {
		return nil, logError("Error reading JSON file", err)
	}

	var result version.Info
	err = json.Unmarshal(byteValue, &result)
	if err != nil {
		return nil, logError("Error unmarshalling JSON", err)
	}
	return &result, nil
}

// logError logs an error message and returns the error.
func logError(message string, err error) error {
	logutil.Warn("%s: %v", message, err)
	return err
}
