// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.
//
// SPDX-License-Identifier: GPL-3.0-or-later

package version

import (
	"errors"
	"regexp"
	"strconv"
	"strings"
	"time"

	"pagure.io/modist/logutil"
)

// Version represents a version with epoch, version, pre-release, commit, timestamp, and
// build.
type Version struct {
	Epoch   int
	Version string
	Pre     string
	Commit  string
	Time    time.Time
	Build   string
}

var (
	ErrUnknownVersionFormat = errors.New("unknown version format")
	ErrBadVersionTime       = errors.New("could not parse the version timestamp")
	ErrUnknownOutputFormat  = errors.New("unknown output format")

	versionRe = regexp.MustCompile(
		`^v?(?P<major>\d+)(\.(?P<minor>\d+)(\.(?P<patch>\d+))?)?$`,
	)
	forbiddenRe         = regexp.MustCompile(`[^0-9A-Za-z.-]+`)
	leadingHyphenRe     = regexp.MustCompile(`-[.-]+`)
	leadingDotRe        = regexp.MustCompile(`\.[.-]+`)
	zeroLeadingNumberRe = regexp.MustCompile(`\.0+([1-9]\d*)\.`)
	hyphenRe            = regexp.MustCompile(`-`)
	versionGr           = `(?P<version>\d+\.\d+\.\d+)`
	timeGr              = `(?P<time>\d{4}[0-1]\d[0-3]\d[0-2]\d[0-5]\d[0-5]\d)` // yyyymmddhhmmss
	preGr               = `(?P<pre>[[:alnum:]]([[:alnum:]-.]*[[:alnum:]])?)`
	commitGr            = `(?P<commit>[[:alnum:]-]{12})`
	buildGr             = `\+(?P<build>[[:alnum:]-.]+)`
	fromGoRe            = regexp.MustCompile(
		`^v` + versionGr + `((-((` + preGr + `\.)?0\.)?` + timeGr + `-` + commitGr + `)|(-` + preGr + `)?)(` + buildGr + `)?$`,
	)
)

// getSubs processes raw with the exp regexp expression, containing named groups;
// returns a map of the matched group values, indexed by the group names.
func getSubs(raw string, exp *regexp.Regexp) (map[string]string, error) {
	result := make(map[string]string)
	match := exp.FindStringSubmatch(raw)
	if match != nil {
		for i, name := range exp.SubexpNames() {
			if i != 0 && name != "" && match[i] != "" {
				result[name] = match[i]
			}
		}
	}
	return result, nil
}

// intString converts a string to an integer and then back to a string.
// This ensures the input string is a valid integer and returns the string representation.
func intString(raw string) (string, error) {
	i, err := strconv.Atoi(raw)
	if err != nil {
		return "", err
	}
	return strconv.Itoa(i), nil
}

// cleanPre cleans up pre-release parts by removing forbidden characters,
// leading hyphens, and dots, and normalizing the string.
func cleanPre(raw string) string {
	result := "." + raw + "."
	result = forbiddenRe.ReplaceAllString(result, "")
	result = leadingHyphenRe.ReplaceAllString(result, "-")
	result = leadingDotRe.ReplaceAllString(result, ".")
	result = zeroLeadingNumberRe.ReplaceAllString(result, ".${1}.")
	result = strings.TrimPrefix(result, ".")
	result = strings.TrimSuffix(result, ".")
	result = strings.TrimSuffix(result, "-")
	result = strings.ToLower(
		result,
	) // Pre-version end up in rpm filenames which should be case-insensitive
	return result
}

// New creates a new Version instance from the provided components.
// It validates and processes each component and returns a Version instance or an error if
// any component is invalid.
func New(version, pre, commit, timestamp, build string, epoch int) (*Version, error) {
	var r Version
	if version != "" {
		vsubs, err := getSubs(version, versionRe)
		if err != nil || len(vsubs) == 0 {
			return nil, ErrUnknownVersionFormat
		}
		major, _ := intString(vsubs["major"])
		minor := "0"
		if vsubs["minor"] != "" {
			minor, _ = intString(vsubs["minor"])
		}
		patch := "0"
		if vsubs["patch"] != "" {
			patch, _ = intString(vsubs["patch"])
		}
		r.Version = major + "." + minor + "." + patch
	}
	if pre != "" {
		r.Pre = cleanPre(pre)
		if r.Pre != "" {
			logutil.Warn("Converting pre-version suffix: %s to “%s”", pre, r.Pre)
		}
	}
	if commit != "" {
		r.Commit = cleanPre(commit)
		if r.Commit != "" {
			logutil.Warn("Converting commit ID: %s to “%s”", commit, r.Commit)
		}
	}
	if timestamp != "" {
		var err error
		r.Time, err = parseTimestamp(timestamp)
		if err != nil {
			return &r, ErrBadVersionTime
		}
	}
	if build != "" {
		r.Build = build
	}
	if epoch != 0 {
		r.Epoch = epoch
	}
	return &r, nil
}

// MergeVersions merges the components of src Version into dst Version,
// only if the respective components in dst are empty or zero.
func MergeVersions(dst, src *Version) {
	if src == nil || dst == nil {
		return
	}
	if src.Epoch != 0 {
		dst.Epoch = src.Epoch
	}
	if src.Version != "" && dst.Version == "" {
		dst.Version = src.Version
	}
	if src.Pre != "" {
		dst.Pre = src.Pre
	}
	if src.Commit != "" {
		dst.Commit = src.Commit
	}
	if !src.Time.IsZero() {
		dst.Time = src.Time
	}
	if src.Build != "" {
		dst.Build = src.Build
	}
}

// parseTimestamp parses a timestamp string into a time.Time instance.
// It supports multiple timestamp formats and returns an error if parsing fails.
func parseTimestamp(timestamp string) (time.Time, error) {
	if timestamp == "" {
		return time.Time{}, nil
	}
	t, err := time.Parse("20060102150405", timestamp)
	if err != nil {
		t, err = time.Parse(time.RFC3339, timestamp)
	}
	if err != nil {
		return t, err
	}
	return t.UTC(), nil
}

// FromGo creates a Version instance from a Go version string.
// It parses the version string and extracts components using regular expressions.
func FromGo(raw string) (*Version, error) {
	vsubs, err := getSubs(raw, fromGoRe)
	if err != nil || len(vsubs) == 0 {
		return nil, ErrUnknownVersionFormat
	}
	return New(
		vsubs["version"],
		vsubs["pre"],
		vsubs["commit"],
		vsubs["time"],
		vsubs["build"],
		0,
	)
}

// Go returns the Go format version string of the Version instance.
// It constructs the string from the version components, following Go versioning
// conventions.
func (ver *Version) Go() string {
	version := ver.Version
	if version == "" {
		version = "0"
	}
	for strings.Count(version, ".") < 2 {
		version += ".0"
	}

	time := ver.Time.Format("20060102150405")
	result := "v" + version
	sep := "-"
	if ver.Pre != "" {
		result += sep + ver.Pre
		sep = "."
	}
	if ver.Commit != "" {
		commit := string([]rune(ver.Commit)[:12])
		sup := time + "-" + commit
		if !strings.HasSuffix(version, ".0.0") {
			sup = "0." + sup
		}
		result += sep + sup
	}

	return result
}

// RPM returns the RPM format version string of the Version instance.
// It constructs the string from the version components, following RPM versioning
// conventions.
func (ver *Version) RPM() string {
	version := ver.Version
	if version == "" {
		version = "0"
	}
	time := ver.Time.Format("20060102")
	result := version
	sep := "^"
	if ver.Pre != "" {
		result += "~" + ver.Pre
	}
	if ver.Commit != "" {
		commit := string([]rune(ver.Commit)[:7])
		result += sep + time + "git" + commit
	}
	result = hyphenRe.ReplaceAllString(result, ".")
	if ver.Epoch > 0 {
		result = strconv.Itoa(ver.Epoch) + ":" + result
	}
	if ver.Build != "" {
		result += "-" + ver.Build
	}
	return result
}

// Format returns the version string in the specified format (e.g., "rpm" or "go").
// It calls the appropriate method to generate the version string or returns an error if the
// format is unknown.
func (ver *Version) Format(format string) (string, error) {
	switch format {
	case "rpm":
		return ver.RPM(), nil
	case "go":
		return ver.Go(), nil
	}
	return "", ErrUnknownOutputFormat
}

// Info represents module proxy information.
type Info struct {
	Version string    // version string, in Go format
	Time    time.Time `json:",omitempty"` // code timestamp (unreliable)
	Build   string    `json:",omitempty"` // Build (semver) / Release (RPM)
	Epoch   int       `json:",omitempty"` // versioning Epoch (RPM)
}

// NewInfo creates a new Info instance with the provided version, time, build, and epoch.
func NewInfo(version string, Time time.Time, build string, epoch int) *Info {
	return &Info{Version: version, Time: Time, Build: build, Epoch: epoch}
}

// Info returns the Info representation of the Version.
// It creates a new Info instance using the version components.
func (ver *Version) Info() *Info {
	return NewInfo(ver.Go(), ver.Time, ver.Build, ver.Epoch)
}

// Ver returns the Version representation of the Info.
// It creates a new Version instance from the Info components and Go version string.
func (ver *Info) Ver() (*Version, error) {
	v, err := FromGo(ver.Version)
	if err != nil {
		return nil, err
	}
	v.Time = ver.Time
	v.Build = ver.Build
	v.Epoch = ver.Epoch
	return v, nil
}
