// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.
//
// SPDX-License-Identifier: GPL-3.0-or-later

package write

import (
	"archive/zip"
	"bytes"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"os"
	"path/filepath"
	"sort"
	"strings"
	"syscall"

	"golang.org/x/mod/module"
	"golang.org/x/mod/sumdb/dirhash"
	"golang.org/x/sys/unix"

	"pagure.io/modist/descriptor"
	"pagure.io/modist/files"
	"pagure.io/modist/logutil"
	"pagure.io/modist/read"
	"pagure.io/modist/version"
)

var (
	ErrWriteNotSafe        = errors.New("destination may already be in use")
	ErrUnknownOutputFormat = errors.New("unknown output format")
)

// Files is a map of produced file paths to a list of special attributes such as "dir" or
// "ghost".
type Files map[string][]string

// File creates a new file.
func (written Files) File(
	reader io.Reader,
	prefix, radical, extension, reference string,
	overwrite bool,
) error {
	realPath, pseudoPath, err := createFilePath(radical, extension, prefix)
	if err != nil {
		return err
	}
	if !overwrite {
		if err := checkFileExists(realPath); err != nil {
			return err
		}
	}
	file, err := createAndWriteFile(realPath, reader)
	if err != nil {
		return err
	}
	defer file.Close()

	if filepath.Ext(pseudoPath) == ".md" {
		written[pseudoPath] = []string{"doc"}
	} else {
		written[pseudoPath] = nil
	}
	return setFileModTime(realPath, reference, file)
}

// createFilePath generates real and pseudo paths based on provided radical, extension, and
// prefix.
func createFilePath(radical, extension, prefix string) (string, string, error) {
	if extension != "" {
		extension = "." + strings.TrimPrefix(extension, ".")
	}
	pseudoPath := strings.TrimSuffix(radical, extension) + extension
	realPath := filepath.Join(prefix, pseudoPath)
	return realPath, pseudoPath, nil
}

// checkFileExists checks if a file already exists at the given path.
func checkFileExists(realPath string) error {
	if _, err := os.Lstat(realPath); !os.IsNotExist(err) {
		return ErrWriteNotSafe
	}
	return nil
}

// createAndWriteFile creates a new file at the specified path and writes the provided data
// to it.
func createAndWriteFile(realPath string, reader io.Reader) (*os.File, error) {
	file, err := os.Create(realPath)
	if err != nil {
		return nil, err
	}
	if _, err = io.Copy(file, reader); err != nil {
		return nil, err
	}
	return file, nil
}

// setFileModTime sets the modification time of the file based on the reference path.
func setFileModTime(realPath, reference string, file *os.File) error {
	if err := SetModTime(realPath, reference); err != nil {
		return err
	}
	return file.Sync()
}

// Dir creates a new directory.
func (written Files) Dir(prefix, path, reference string) error {
	realPath := filepath.Join(prefix, path)
	if err := checkDirExists(realPath); err != nil {
		logutil.Info("Directory %s already exists!\n", realPath)
	}
	if err := createDir(realPath); err != nil {
		return err
	}
	written[path] = []string{"dir"}
	return SetModTime(realPath, reference)
}

// checkDirExists checks if a directory already exists at the given path.
func checkDirExists(realPath string) error {
	if _, err := os.Lstat(realPath); !os.IsNotExist(err) {
		return err
	}
	return nil
}

// createDir creates a new directory at the specified path.
func createDir(realPath string) error {
	return os.MkdirAll(realPath, 0o777)
}

// Symlink creates a new symbolic link.
func (written Files) Symlink(prefix, path, target, reference string) error {
	realPath := filepath.Join(prefix, path)
	if err := checkFileExists(realPath); err != nil {
		return err
	}
	if err := os.Symlink(target, realPath); err != nil {
		return err
	}
	written[path] = nil
	return SetModTime(realPath, reference)
}

// Descriptor creates a new packed module description file.
func (written Files) Descriptor(
	descriptor *descriptor.Descriptor,
	prefix, radical, reference string,
) error {
	if err := descriptor.Trim(); err != nil {
		return err
	}
	data, err := descriptor.Format()
	if err != nil {
		return err
	}
	return written.File(bytes.NewReader(data), prefix, radical, "mod", reference, false)
}

// Info creates a new module info file.
func (written Files) Info(info *version.Version, prefix, radical, reference string) error {
	data, err := json.MarshalIndent(info.Info(), "", "  ")
	if err != nil {
		return fmt.Errorf(
			"failed to marshal module revision information as JSON data: %v",
			err,
		)
	}
	return written.File(bytes.NewReader(data), prefix, radical, "info", reference, false)
}

// Zip creates a new packed module zip file.
func (written Files) Zip(module *files.Files, prefix, radical string) error {
	reader, writer := io.Pipe()
	defer reader.Close()
	defer writer.Close()
	errors := make(chan error)
	go func() {
		errors <- written.File(reader, prefix, radical, "zip", "", false)
	}()
	go func() {
		err := zipFiles(writer, module)
		writer.Close()
		errors <- err
	}()
	for i := 0; i < 2; i++ {
		if err := <-errors; err != nil {
			return err
		}
	}
	return nil
}

// Hash creates a new packed module hash file.
func (written Files) Hash(prefix, radical string) error {
	pseudoZipPath := strings.TrimSuffix(radical, ".ziphash") + ".zip"
	realZipPath := filepath.Join(prefix, pseudoZipPath)
	data, err := dirhash.HashZip(realZipPath, dirhash.DefaultHash)
	if err != nil {
		return fmt.Errorf(
			"could not compute the hash of “%s” (packed module zip file): %v",
			realZipPath,
			err,
		)
	}
	return written.File(strings.NewReader(data+"\n"), prefix, radical, "ziphash", "", false)
}

// Pack creates a packed module.
func (written Files) Pack(mod *files.Files, prefix, destination string) error {
	encPath, encVer, err := escapeModulePathAndVersion(mod)
	if err != nil {
		return err
	}
	relDestination := filepath.Join(encPath, "@v")
	fullDestination := filepath.Join(destination, relDestination)
	if err := createAllDirs(filepath.Join(prefix, fullDestination)); err != nil {
		return err
	}
	recordDirs(written, relDestination, destination)
	radical := filepath.Join(fullDestination, encVer)
	return packFiles(written, mod, prefix, radical, fullDestination)
}

// escapeModulePathAndVersion escapes the module path and version for use in file paths.
func escapeModulePathAndVersion(mod *files.Files) (string, string, error) {
	path := mod.Descriptor.Module.Mod.Path
	encPath, err := module.EscapePath(path)
	if err != nil {
		return "", "", err
	}
	encVer, err := module.EscapeVersion(mod.Version.Go())
	if err != nil {
		return "", "", err
	}
	return encPath, encVer, nil
}

// createAllDirs creates all directories along the given path.
func createAllDirs(path string) error {
	return os.MkdirAll(path, 0o777)
}

// recordDirs records directory paths in the written Files map.
func recordDirs(written Files, relDestination, destination string) {
	for relDestination != "." {
		written[filepath.Join(destination, relDestination)] = []string{"dir"}
		relDestination = filepath.Dir(relDestination)
	}
}

// packFiles handles the creation of zip, hash, descriptor, and info files for the module.
func packFiles(
	written Files,
	mod *files.Files,
	prefix, radical, fullDestination string,
) error {
	if err := written.Zip(mod, prefix, radical); err != nil {
		return err
	}
	if err := written.Hash(prefix, radical); err != nil {
		return err
	}
	if err := written.Descriptor(mod.Descriptor, prefix, radical, ""); err != nil {
		return err
	}
	if err := written.Info(mod.Version, prefix, radical, ""); err != nil {
		return err
	}
	written[filepath.Join(fullDestination, "list")] = []string{"ghost"}
	return nil
}

// SetModTime sets the modification time of path to the same as reference.
func SetModTime(path, reference string) error {
	if reference != "" {
		info, err := os.Lstat(reference)
		if os.IsNotExist(err) {
			return nil
		} else if err != nil {
			return err
		}
		stat := info.Sys().(*syscall.Stat_t)
		atime := unix.NsecToTimeval(syscall.TimespecToNsec(stat.Atim))
		mtime := unix.NsecToTimeval(syscall.TimespecToNsec(stat.Mtim))
		return unix.Lutimes(path, []unix.Timeval{atime, mtime})
	}
	return nil
}

// Text returns a text representation of the written files.
func (written Files) Text(prefix string, special bool) string {
	paths := getSortedPaths(written)
	var data string
	for _, path := range paths {
		if written[path] == nil || special {
			data += filepath.Join(prefix, path) + "\n"
		}
	}
	return data
}

// getSortedPaths returns a sorted list of paths from the written Files map.
func getSortedPaths(written Files) []string {
	var paths []string
	for path := range written {
		paths = append(paths, path)
	}
	sort.Strings(paths)
	return paths
}

// RPM returns an RPM representation of the written files.
func (written Files) RPM() string {
	paths := getSortedPaths(written)
	var data string
	for _, path := range paths {
		for _, mod := range written[path] {
			data += "%" + mod + " "
		}
		data += path + "\n"
	}
	return data
}

// JSON returns a JSON representation of the written files.
func (written Files) JSON() (string, error) {
	data, err := json.MarshalIndent(written, "", "  ")
	if err != nil {
		return "", fmt.Errorf("failed to marshal results as JSON data: %v", err)
	}
	return string(data), nil
}

// Format returns the formatted representation of the written files.
func (written Files) Format(format string) (string, error) {
	switch format {
	case "json":
		return written.JSON()
	case "rpm":
		return written.RPM(), nil
	case "txt":
		return written.Text("", true), nil
	default:
		return "", ErrUnknownOutputFormat
	}
}

// Output writes the formatted representation of the written files to a file.
func (written Files) Output(file, format string) error {
	data, err := written.Format(format)
	if err != nil {
		return err
	}
	ext := format
	if format == "rpm" {
		ext = "lst"
	}
	return written.File(strings.NewReader(data), "", file, ext, "", true)
}

// zipFiles adds files to a zip archive.
func zipFiles(writer io.Writer, module *files.Files) error {
	zipWriter := zip.NewWriter(writer)
	defer zipWriter.Close()
	paths := make(map[string]bool)
	for _, path := range module.Selected {
		paths[path] = true
	}
	dirs := make(map[string]bool)
	for _, path := range module.Selected {
		dirs[filepath.Dir(path)] = true
	}
	delete(dirs, ".")
	for dir := range dirs {
		for root := filepath.Dir(dir); root != "."; root = filepath.Dir(root) {
			dirs[root] = true
		}
	}
	for _, path := range module.Selected {
		err := addFileToZip(zipWriter, module, path, paths, dirs)
		if err != nil {
			return err
		}
	}
	return nil
}

// addFileToZip adds a single file to the zip archive.
func addFileToZip(
	zipWriter *zip.Writer,
	module *files.Files,
	path string,
	paths, dirs map[string]bool,
) error {
	spath := filepath.Join(module.Root, path)
	info, err := os.Lstat(spath)
	if err != nil {
		return err
	}
	target, mode, err := getFileModeAndTarget(module, path, spath, paths, dirs)
	if err != nil {
		return err
	}
	header, err := zip.FileInfoHeader(info)
	if err != nil {
		return err
	}
	header.Name = filepath.Join(
		module.Descriptor.Module.Mod.Path+"@"+module.Version.Go(),
		path,
	)
	header.Method = zip.Deflate
	header.SetMode(mode)
	writer, err := zipWriter.CreateHeader(header)
	if err != nil {
		return err
	}
	return writeFileToZip(writer, spath, target)
}

// getFileModeAndTarget determines the file mode and target path for a file to be added to
// the zip archive.
func getFileModeAndTarget(
	module *files.Files,
	path, spath string,
	paths, dirs map[string]bool,
) (string, os.FileMode, error) {
	target, mt, st, err := read.Symlink(module.Root, path)
	switch err {
	case read.ErrNotASymLink:
		return "", 0o644, nil
	case read.ErrOutsideSymlink:
		logutil.Warn("%s target %s is outside the module\n", path, target)
		return target, 0o755 | os.ModeSymlink, nil
	case nil:
		if target != st {
			logutil.Info("Simplifying %s target %s to %s\n", path, target, st)
		}
		if !isPathIncluded(mt, paths, dirs) {
			logutil.Warn("%s target %s is not included in the module\n", path, st)
		}
		return st, 0o755 | os.ModeSymlink, nil
	default:
		return "", 0, err
	}
}

// isPathIncluded checks if the given path is included in the module paths or directories.
func isPathIncluded(mt string, paths, dirs map[string]bool) bool {
	_, existsP := paths[mt]
	_, existsD := dirs[mt]
	return existsP || existsD
}

// writeFileToZip writes the file data to the zip archive.
func writeFileToZip(writer io.Writer, spath, target string) error {
	if target == "" {
		file, err := os.Open(spath)
		if err != nil {
			return err
		}
		defer file.Close()
		_, err = io.Copy(writer, file)
		return err
	}
	_, err := io.Copy(writer, strings.NewReader(target))
	return err
}
